import os
import tkinter as tk
import tkinter.messagebox
from tkinter import ACTIVE, LEFT, RIGHT, Button, Frame, filedialog, ttk
from tkinter.simpledialog import Dialog

from .uebase import desktop

"""This module contains the gui elements used by uvs, the UnrealVersionSelector"""


def find_icon():
    # we need to find an icon for the handler
    paxoicon = os.path.abspath(os.path.join(os.path.dirname(__file__), "paxo.ico"))
    if not os.path.isfile(paxoicon):
        return None
    return paxoicon


class EngineSelector(Dialog):
    def __init__(self, engines, initial_id=None):
        title = "Select Unreal Engine Version"
        self.initial_id = initial_id
        self.filepicker_id = None
        self.result = None
        self.engines = engines
        self.descr = {
            k: desktop.get_engine_description(k, v) for k, v in engines.items()
        }
        self.ids = list(engines.keys())

        def compare_ids(a, b):
            if a == b:
                return 0
            if desktop.is_preferred_identifier(a, b):
                return -1
            return 1

        self.ids = desktop.sort_identifiers(self.ids)
        try:
            self.idx = self.ids.index(initial_id)
        except ValueError:
            self.idx = -1
        super().__init__(None, title)

    def body(self, master):
        self.minsize(450, 0)  # Set the minimum size of the dialog
        self.attributes("-topmost", True)  # Set the dialog to be on top
        # self.attributes("-toolwindow", True)  # Set the dialog to be a tool window (no minimize/maximize buttons
        self.iconbitmap(find_icon())
        # grow the master frame to fill the window
        master.pack(fill=tk.BOTH, expand=True)

        # Create a Frame for the body
        # body_frame = tk.Frame(master)
        # body_frame.pack(fill=tk.BOTH, expand=True)  # Set fill and expand options

        # Create a Combobox for the left side
        values = [self.descr[k] for k in self.ids]
        combobox = ttk.Combobox(master, values=values)
        if self.idx >= 0:
            combobox.current(self.idx)
        combobox.pack(
            side=tk.LEFT, fill=tk.BOTH, expand=True
        )  # Set fill and expand options
        self.combobox = combobox

        # Create a Button for the right side
        button = tk.Button(master, text=" ... ", command=self.filepicker)
        button.pack(side=tk.RIGHT)  # Set fill and expand options

    def buttonbox(self):
        """add standard button box.

        overriding to place buttons to the right
        """

        box = Frame(self)

        w = Button(box, text="OK", width=10, command=self.ok, default=ACTIVE)
        w.pack(side=LEFT, padx=5, pady=5)
        w = Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

        # pack the box to the right
        box.pack(side=RIGHT)

    def filepicker(self):
        # find the first engine root which is not a x.x id
        for id in self.ids:
            if not desktop.is_stock_engine_release(id):
                start = self.engines[id]
                break
        else:
            start = None

        engine_root = filedialog.askdirectory(
            parent=self,
            initialdir=start,
            title="Select the Unreal Engine installation to use for this project",
        )
        if not engine_root:
            return
        if not desktop.is_valid_root_directory(engine_root):
            tkinter.messagebox.showerror(
                parent=self,
                title="Error",
                message=f"The selected directory is not a valid engine installation: {engine_root}",
            )
            return  # try again

        self.filepicker_id = desktop.get_engine_identifier_from_root_dir(engine_root)
        self.ok()

    def apply(self):
        if self.filepicker_id:
            self.result = self.filepicker_id
        else:
            idx = self.combobox.current()
            if idx >= 0:
                self.result = self.ids[idx]


class ProjectFilesWindow(desktop.FeedbackSink):
    def __init__(self):
        self.cancelled = False
        self.root = tk.Tk()
        self.root.title("Generate Project Files")
        self.root.iconbitmap(find_icon())
        self.root.attributes("-topmost", True)  # Set the dialog to be on top
        self.root.protocol("WM_DELETE_WINDOW", self.on_close)

        self.upper_frame = tk.Frame(self.root)
        self.upper_frame.configure(width=400, height=50)
        self.upper_frame.pack(fill=tk.BOTH, expand=False)
        self.lower_frame = tk.Frame(self.root, relief=tk.RAISED, borderwidth=1)
        self.lower_frame.pack(fill=tk.BOTH, expand=True)

        progressframe = tk.Frame(self.upper_frame)
        self.label = tk.Label(progressframe, text="Generating project files...")
        self.label.pack(expand=True, pady=10, anchor="w")

        self.progressbar = ttk.Progressbar(
            progressframe, orient="horizontal", length=300, mode="determinate"
        )
        self.progressbar.pack(pady=10)
        progressframe.pack(side="left", expand=False, padx=5)

        # add a button to show the log window
        self.show_log_button = tk.Button(
            self.upper_frame, text="Show Log", command=self.show_log
        )
        self.show_log_button.pack(side="right", padx=5)

        self.logwindow = tk.Text(self.lower_frame, height=10, wrap="word")
        self.logwindow.pack(side=tk.LEFT, fill=tk.BOTH, expand=True)
        scroll = tk.Scrollbar(self.lower_frame)
        scroll.pack(side=tk.RIGHT, fill=tk.Y)
        self.logwindow.config(yscrollcommand=scroll.set)
        scroll.config(command=self.logwindow.yview)
        self.lower_frame.pack_forget()

    def show_log(self):
        """Shows the log window"""
        if not self.lower_frame.winfo_ismapped():
            self.lower_frame.pack(fill=tk.BOTH, expand=True)
            self.show_log_button.config(text="Hide Log")
        else:
            self.lower_frame.pack_forget()
            self.show_log_button.config(text="Show Log")

    def add_log_line(self, line: str):
        """Adds a line to the text in the log window"""
        if not self.root:
            return
        self.logwindow.insert(tk.END, line + "\n")
        # scroll to the end
        self.logwindow.see(tk.END)

    def update_progress_bar(self, progress: float):
        """Updates the progress bar to the given value"""
        # progress is a flat in the range 0..1
        self.progressbar["value"] = progress * 100

    def update_progress_bar_title(self, title: str):
        """Updates the progress bar title"""
        self.label["text"] = title

    def on_close(self):
        self.cancel()

    def cancel(self):
        """Just signal that the operation should be cancelled"""
        self.cancelled = True
        self.root.quit()

    def close(self):
        self.root.destroy()

    # FeedbackSink methods
    def poll(self, delay: float = 0.0) -> str:
        """Polls the feedback sink for any new messages and runs the event loop"""
        # if window isn't destroyed, run the event loop
        if self.cancelled:
            return "cancel"
        if delay > 0.0:
            self.root.after(int(delay * 1000), self.root.quit)
            self.root.mainloop()
        else:
            self.root.update()
        if self.cancelled:
            return "cancel"
        return ""

    def log(self, message: str, level: str = "info") -> None:
        """Logs a message to the feedback sink"""
        self.add_log_line(message)

    def progress_message(self, message: str) -> None:
        self.update_progress_bar_title(message)

    def progress_value(self, progress: float) -> None:
        self.update_progress_bar(progress)

    @classmethod
    def test(cls):
        """Test the ProjectFilesWindow"""
        window = cls()

        line = "This is a test line"
        i = 0

        def update():
            nonlocal i
            window.add_log_line(f"{line} {i}")
            i += 1
            window.root.after(1000, update)

        window.root.after(100, update)
        window.root.mainloop()


if __name__ == "__main__":
    ProjectFilesWindow.test()



# some features of the ue paths thing
from . import platform

_engine_root: str = "C:/Program Files/Epic Games/UE_4.27/Engine/"
_project_root: str = "C:/Users/username/Documents/Unreal Projects/MyProject/"


def set_engine_root(path: str):
    """Set the engine root path."""
    global _engine_root
    _engine_root = path
    if not _engine_root.endswith("/"):
        _engine_root += "/"
    platform.reset()


def set_project_root(path: str):
    """Set the project root path."""
    global _project_root
    _project_root = path
    if not _project_root.endswith("/"):
        _project_root += "/"
    platform.reset()


def engine_dir() -> str:
    """We are not an engine.  Just return a wonko string."""
    return _engine_root + "Engine/"


def project_dir() -> str:
    """We are not a project.  Just return a wonko string."""
    return _project_root


def engine_config_dir():
    return engine_dir() + "Config/"


def project_config_dir():
    return project_dir() + "Config/"


def engine_patform_extensions_dir() -> str:
    return engine_dir() + "Platforms/"


def project_patform_extensions_dir() -> str:
    return project_dir() + "Platforms/"



import os
import os.path
import re
import subprocess
import sys

import click

# Install / upgrade python in the global environment
# Yes, I know, we must be already running python, but... here we can upgrade it or install
# new versions, etc.

major, minor = sys.version_info[:2]
this_version = f"{major}.{minor}"


def validate_version(version):
    parts = version.split(".")
    if len(parts) < 2 or len(parts) > 3:
        raise ValueError(f"Invalid version {version}")
    for part in parts:
        if not part.isdigit():
            raise ValueError(f"Invalid version {version}")


def get_major(version):
    return ".".join(version.split(".")[:2])


def install_python(version="3.12"):
    """Install python in the global environment."""
    validate_version(version)

    # check if it is already installed
    if is_installed(version):
        print(f"Python {version} already installed")
        return

    installdir = get_install_location(version)

    # install via winget
    print(f"Installing python {version}...")
    subprocess.run(
        [
            "winget",
            "install",
            "--accept-source-agreements",
            "--accept-package-agreements",
            "--scope",
            "machine",
            "-e",
            "--id",
            f"Python.Python.{major}",
            "--location",
            installdir,
        ],
        check=True,
    )
    print("Python installed")


def get_install_location(version):
    """Get the installation location of python version."""
    # create something like C:\Python312
    major = get_major(version)
    program_files = os.environ["ProgramFiles"]
    drive = os.path.splitdrive(program_files)[0]
    suffix = "".join(major.split("."))
    installdir = os.path.join(drive, "\\Python" + suffix)
    return installdir


def upgrade_python(
    version=None, silent=False, interactive=False, install=False, self_upgrade=False
):
    """Upgrade python in the global environment."""
    if version is None:
        version = get_version()
    major = get_major(version)

    if major == this_version and not self_upgrade:
        print(
            f"Cannot upgrade python {version} from itself, use --self-upgrade to create a bat file to do it."
        )
        return False

    # check if it is already installed
    if not is_installed(version):
        if install:
            return install_python(version, silent=silent, interactive=interactive)
        else:
            raise ValueError(f"Python {version} is not installed")

    # winget upgrade is not working, it won't accept the machine scope and will put the new version
    # in the default location.  Instead, use install, which will upgrade if it can, but we must
    # use 'force' to force winget to download the installer.

    # get current install dir
    installdir = get_install_dir(major)

    # install via winget
    cmd = [
        "winget",
        "install",
        "--force",  # force download of installer
        "--accept-source-agreements",
        "--accept-package-agreements",
        "-e",
        "--id",
        f"Python.Python.{major}",
        "--location",
        installdir,
        "--scope",
        "machine",
    ]
    if silent:
        cmd.append("--silent")
    if interactive:
        cmd.append("--interactive")

    if major != this_version:
        print(f"Upgrading python {version}...")
        subprocess.run(cmd, check=True)
        print("Python upgraded")
    else:
        print(f"Creating bat file to upgrade python {version}...")
        filename = "upgrade_python.bat"
        with open(filename, "w") as f:
            f.write(" ".join(cmd))
        print(f"run '{filename}' to upgrade python.")
    return True


def get_version(major=None):
    """Get the version of python installed in the global environment."""
    if major is None:
        cmd = ["py", "--version"]
    else:
        cmd = ["py", f"-{major}", "--version"]
    out = subprocess.run(cmd, capture_output=True)
    if out.returncode != 0:
        return None  # not installed
    return out.stdout.decode().strip().split()[1]


def is_installed(version):
    """Check if python version is installed"""

    major = get_major(version)
    out = subprocess.run(["py", f"-{major}", "--version"], capture_output=True)
    return out.returncode == 0


def get_install_dir(version):
    """Get the installation directory of python version."""
    major = get_major(version)
    out = subprocess.run(["py", "--list-paths"], capture_output=True, check=True)
    for line in out.stdout.decode().splitlines():
        prefix, path = line.rsplit(None, 1)
        match = re.search(r"-V:(\d+\.\d+)", prefix)
        if match:
            version = match.group(1)
            if version == major:
                location = os.path.normpath(os.path.dirname(path))
                return location


@click.group()
@click.option("--interactive", is_flag=True, help="Requests interactive mode.")
@click.option("--silent", is_flag=True, help="Requests silent mode.")
@click.pass_context
def python(ctx, interactive, silent):
    """Manage python in the global environment."""
    ctx.ensure_object(dict)
    ctx.obj["interactive"] = interactive
    ctx.obj["silent"] = silent


@python.command()
@click.option("--version", "-v", default="3.12", help="Version to install.")
@click.pass_context
def install(ctx, version):
    """Install python in the global environment."""
    install_python(
        version, interactive=ctx.obj["interactive"], silent=ctx.obj["silent"]
    )


@python.command()
@click.option("--version", "-v", type=str, default=None, help="Version to install.")
@click.option(
    "--self-upgrade",
    is_flag=True,
    help="Create a bat file to upgrade python from itself.",
)
@click.pass_context
def upgrade(ctx, version, self_upgrade):
    """Upgrade python in the global environment."""
    upgrade_python(
        version,
        interactive=ctx.obj["interactive"],
        silent=ctx.obj["silent"],
        self_upgrade=self_upgrade,
    )



import datetime
import json
import os.path
import subprocess
import sys
import tkinter as tk
import tkinter.messagebox
from typing import List, Optional

import click

from . import tools, ue
from .uebase import desktop
from .uvs_gui import EngineSelector, ProjectFilesWindow

# our custom version of UnrealVersionSelector.exe

use_gui = sys.executable.endswith("pythonw.exe")


def find_icon():
    # we need to find an icon for the handler
    paxoicon = os.path.abspath(os.path.join(os.path.dirname(__file__), "paxo.ico"))
    if not os.path.isfile(paxoicon):
        return None
    return paxoicon


def abort(msg):
    raise click.ClickException(msg)


@click.group(invoke_without_command=False)
@click.option("--gui/--no-gui", default=use_gui, help="run with or without gui")
@click.pass_context
def uvs(ctx, gui):
    """A python version of the UnrealVersionSelector.exe.

    Initial command can be prefixed with a single dash or slash."""
    global use_gui
    use_gui = gui

    if ctx.invoked_subcommand is None:
        # if this were part of engine, we could register _this_ engine with prompt
        # but since it's not, we don't support this functionality
        # of invoking UnrealVersionSelector.exe with no arguments.
        pass


cli = uvs


@cli.command()
@click.option("--unattended", is_flag=True, help="Don't prompt for input")
@click.argument("engine_id", default="")
@click.option(
    "--engine-root",
    type=click.Path(exists=True),
    help="The engine root to register",
)
def register(engine_id, unattended, engine_root):
    """Register a _this_ engine.  Not supported."""
    register_current_engine_directory(engine_id, unattended, engine_root)


def register_current_engine_directory(
    engine_id=None, unattended=False, engine_root=None
):
    if not engine_root:
        abort("can't register _this_engine, because this is a standalone tool")
    engine_root = os.path.abspath(engine_root)
    if not desktop.is_valid_root_directory(engine_root):
        abort(f"Invalid engine root {engine_root}")

    if not engine_id:
        engine_id = desktop.get_engine_identifier_from_root_dir(engine_root)
        if not engine_id:
            abort(f"Can't find engine identifier for {engine_root}")

    desktop.platform.register_engine_installation_identifier(engine_root, engine_id)

    # check if file handlers are in place
    if not desktop.platform.verify_file_associations(
        user=True, silent=True
    ) and not desktop.platform.verify_file_associations(user=False, silent=True):
        if unattended:
            do_it = True
        else:
            do_it = messagebox_yesno(
                msg="Register Unreal Engine file types?", heading="File Types"
            )
        if do_it:
            # okay, we update it.
            # we use the engine's own UnrealVersionSelector if present, else we use this script
            desktop.platform.update_file_associations(
                engine_root=engine_root, handler=find_run_args(gui=True)
            )

    click.echo(f"Registered engine {engine_id} at {engine_root}")


# flags are non-standard extensions to the original UnrealVersionSelector.exe
@cli.command()
@click.option("--check", is_flag=True, help="Check file associations.")
@click.option("--clear", is_flag=True, help="Remove associations.")
@click.option("--user", is_flag=True, help="use 'user' scope instead of 'machine'")
@click.option(
    "--engine-root",
    type=click.Path(exists=True),
    help="The engine root for the UnrealVersionSelector.exe to use",
)
def fileassociations(check, clear, user, engine_root):
    """Register file associations"""

    update_file_associations(check, clear, user, engine_root)


def update_file_associations(check=False, clear=False, user=True, engine_root=None):
    if check:
        handler = desktop.platform.verify_file_associations(user=user)
        click.echo(
            f"File associations for scope {'user' if user else 'machine'}: {handler!r}"
        )
    elif clear:
        desktop.platform.clear_file_associations(user=user)
        click.echo(
            f"File associations cleared for scope {'user' if user else 'machine'}"
        )
    else:
        if engine_root:
            desktop.platform.update_file_associations(
                engine_root=engine_root, verify=True
            )
        else:
            handler = find_run_args(gui=True)
            desktop.platform.update_file_associations(handler=handler, verify=True)
        click.echo(
            f"File associations updated for scope {'user' if user else 'machine'}"
        )


# non-standard extension to the original UnrealVersionSelector.exe
@cli.command()
@click.argument(
    "uproject", type=click.Path(exists=True), default=None, metavar="<.uproject>"
)
@click.option(
    "--engine-root",
    type=click.Path(exists=True),
    help="The engine root to return the identifier for",
)
def identifier(uproject, engine_root):
    """Print the identifier for the project"""
    if engine_root:
        identifier = desktop.get_engine_identifier_from_root_dir(engine_root)
    else:
        identifier = desktop.get_engine_identifier_for_project(uproject)
    messagebox_ok(identifier, True, "Engine Identifier")


# non-standard extension to the original UnrealVersionSelector.exe
@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def engineroot(uproject):
    """Print the engine root for the project"""
    root = get_project_engine(uproject)
    messagebox_ok(root, True, "Engine Root")


def get_project_engine(uproject):
    engine_id = desktop.get_engine_identifier_for_project(uproject)
    if not engine_id:
        raise click.ClickException(f"Can't find engine identifier for {uproject}")

    engine_root = desktop.get_engine_root_dir_from_identifier(engine_id)
    if not engine_root:
        raise click.ClickException(f"Can't find engine root for {engine_id}")
    return engine_root


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
@click.option(
    "--engine_id", metavar="<engine_id>", default=None, help="The engine id to use"
)
def switchversion(uproject, engine_id):
    """Switch the engine version for the project"""

    current_engine_id = desktop.get_engine_identifier_for_project(uproject)
    engines = desktop.platform.enumerate_engine_installations()
    if not engine_id:
        engine_id = select_engine_installation(engines, initial_id=current_engine_id)
    elif engine_id not in engines:
        abort(f"Unknown engine id {engine_id}")

    if not desktop.set_engine_identifier_for_project(uproject, engine_id):
        abort(
            f"Failed to set engine id {engine_id} for {uproject}.  Check if the file is writeable."
        )

    # if it is a content only project, we are done
    status = desktop.query_status_for_project(uproject)
    if not status.code_based_project:
        return

    # Need to rebuild project files
    return generate_project_files(uproject)


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def editor(uproject):
    """Start the editor for project"""
    return launch_editor(uproject)


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def projectlist(uproject):
    """Start the editor with a list of projects"""
    return launch_editor()


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def game(uproject):
    """Start the editor for the project and run the game"""
    return launch_editor(uproject, ["-game"])


@cli.command()
@click.argument("uproject", type=click.Path(exists=True), metavar="<.uproject>")
def projectfiles(uproject):
    """Rebuild project files for the project"""
    # editor = ue.get_editor_from_uproject()
    return generate_project_files(uproject)


def launch_editor(uproject: Optional[str], args: List[str] = []):
    """Start the editor for project"""

    if not uproject:
        engines = desktop.platform.enumerate_engine_installations()
        engine_id = select_engine_installation(engines)
        if not engine_id:
            return False
        root_dir = desktop.get_engine_root_dir_from_identifier(engine_id)
        if not platform_launch_editor(root_dir, None, args):
            abort("Failed to launch editor")
        return True

    root = get_validated_engine_root(uproject)
    if not root:
        return False
    # Figure out the path to the editor executable. This may be empty for older .target files
    editor_filename = try_get_editor_filename(os.path.join(root, "Engine"), uproject)

    if not platform_launch_editor(root, editor_filename, [uproject] + args):
        abort("Failed to launch editor")
    return True


def get_validated_engine_root(uproject: str) -> Optional[str]:
    root = get_engine_root_dir_for_project(uproject)
    if not root:
        if not switchversion(uproject, None):
            return None
    root = get_engine_root_dir_for_project(uproject)
    if not root:
        abort("Error retrieving project root directory")
    return root


def get_engine_root_dir_for_project(uproject: str) -> Optional[str]:
    engine_id = desktop.get_engine_identifier_for_project(uproject)
    if not engine_id:
        return None
    return desktop.get_engine_root_dir_from_identifier(engine_id)


def try_get_editor_filename(engine_engine: str, uproject: str) -> Optional[str]:
    """Try to find the editor filename for the uproject"""
    project_dir = os.path.dirname(uproject)
    binaries_dir = os.path.join(project_dir, "Binaries", "Win64")
    if os.path.isdir(binaries_dir):
        # find all .target files in the binaries directory and sort them by modification time
        target_files = []
        for root, dirs, files in os.walk(binaries_dir):
            del dirs[:]
            for file in files:
                if file.endswith(".target"):
                    f = os.path.join(root, file)
                    target_files.append((f, os.path.getmtime(f)))
        target_files.sort(key=lambda x: x[1], reverse=True)
        # look for launch path in the target files, last modification first.
        for target_file, _ in target_files:
            launch_path = read_launch_path_from_target_file(target_file)
            if launch_path:
                return launch_path
    return None


def read_launch_path_from_target_file(target_file: str) -> Optional[str]:
    """Read the launch path from the target file"""
    with open(target_file, "r") as f:
        try:
            info = json.load(f)
        except json.JSONDecodeError:
            return None

    # Check it's an editor target
    if info.get("TargetType") != "Editor":
        return None
    # Check it's development configuration
    if info.get("Configuration") != "Development":
        return None
    return info.get("Launch", None)


def platform_launch_editor(
    root_dir: Optional[str], explicit_filename: Optional[str], args: List[str]
) -> bool:
    """Launch the editor for the project"""
    if not explicit_filename:
        assert root_dir
        filename = explicit_filename or os.path.join(
            root_dir, "Engine", "Binaries", "Win64", "UnrealEditor.exe"
        )
    else:
        filename = explicit_filename

    try:
        subprocess.Popen([filename] + args)
    except OSError:
        return False
    return True


def generate_project_files(uproject):
    # check if it is a code project
    # hm, why not use ProjectDescriptor?
    # project = ProjectDescriptor.load(uproject)
    source_dir = os.path.join(os.path.dirname(uproject), "Source")
    if not os.path.isdir(source_dir):
        abort(
            "This project does not have any source code. You need to add C++ source files to the project from the Editor before you can generate project files."
        )

    root_dir = get_validated_engine_root(uproject)
    if not root_dir:
        return False

    exename = "paxo"
    now = datetime.datetime.now().strftime("%Y.%m.%d-%H.%M.%S")
    logfile = os.path.join(
        os.path.dirname(uproject), "Saved", "Logs", f"{exename}-{now}.log"
    )

    if use_gui:
        window = ProjectFilesWindow()
        context = desktop.SinkFeedbackContext(sink=window)
    else:
        window = None
        context = desktop.ClickFeedbackContext()
    try:
        ok = desktop.generate_project_files(root_dir, uproject, context, logfile)
        if not ok:
            log = context.get_logs()
            messagebox_ok(log, False, "Failed to generate project files.")
            return False
        return True
    finally:
        if window:
            window.close()


# @cli.command()
def register_test():
    """Register uvs as the default UnrealVersionSelector"""
    print(sys.argv)
    print(__spec__)
    print(sys.executable)
    return
    # find the command used to run us
    cmd = sys.argv[:]
    i = cmd.index("register-test")
    cmd = cmd[:i]
    # remove any trailing args
    while cmd[-1].startswith("-"):
        cmd.pop()

    # if we are running a .py file, we need to use the -m flag with the actual
    # module name (not __main__)
    if cmd[0].endswith(".py"):
        exe = sys.executable
        if exe.endswith("python.exe"):
            exe = exe[:-10] + "pythonw.exe"
        cmd = [exe, "-m", __spec__.name] + cmd[1:]

    ue.register_uproject_handler(engine_path=None, handler=cmd, test=True)


def find_run_args(gui: bool = False):
    """Find the command to run this script using python"""
    exe = sys.executable
    if gui:
        # on windows, find the pythonw.exe if it exists
        base = os.path.basename(exe)
        if os.path.normcase(base) == "python.exe":
            exew = os.path.join(os.path.dirname(exe), "pythonw.exe")
            if os.path.isfile(exew):
                exe = exew
    return [exe, "-m", __spec__.name]  # type: ignore[name-defined]


def messagebox_ok(
    message: str,
    success: bool = True,
    heading: Optional[str] = None,
    parent: Optional[tk.Tk] = None,
) -> bool:
    if not heading:
        heading = "Success" if success else "Error"

    if use_gui:
        if success:
            tkinter.messagebox.showinfo(heading, message, parent=parent)
        else:
            tkinter.messagebox.showerror(heading, message, parent=parent)
    else:
        # just use echo
        if success:
            click.echo(message)
        else:
            click.echo(
                click.style(f"{heading}:", fg="red") + f" {message}", file=sys.stderr
            )

    return success


def messagebox_yesno(message: str, heading: Optional[str] = None) -> bool:
    heading = heading or "I say:"
    if use_gui:
        return tkinter.messagebox.askyesno(heading, message)
    else:
        return click.confirm(message)


def select_engine_installation(engines, initial_id=None):
    if use_gui:
        d = EngineSelector(engines, initial_id)
        return d.result
    else:
        # prompt for the engine root to use using click
        click.echo("Select Engine.  Known engines:")
        keys = desktop.sort_identifiers(engines.keys())
        for i, k in enumerate(keys):
            v = engines[k]
            description = desktop.get_engine_description(k, v, format=True)
            click.echo(
                f"{click.style(f'{i+1}', fg='yellow', reverse=True)}: {description}"
            )
        try:
            initial_idx = keys.index(initial_id) + 1
        except ValueError:
            initial_idx = None
        while True:
            if len(keys):
                prompt = f"Select engine (1 to {len(keys)}) or enter path to engine"
            else:
                prompt = "Enter engine path to engine"
            result = click.prompt(prompt, type=str, default=initial_idx)
            try:
                result = int(result)
                if 1 <= result <= len(keys):
                    return keys[result - 1]
                continue
            except ValueError:
                pass
            if not os.path.exists(result):
                click.echo(
                    f"{click.style('Error:', fg='red')} Path {click.style(result, fg='green')} not found."
                )
                continue
            if desktop.is_valid_root_directory(result):
                return desktop.get_engine_identifier_from_root_dir(result)
            click.echo(
                f"{click.style('Error:', fg='red')} Path {click.style(result, fg='green')} does not contain a valid Unreal Engine."
            )


def main():
    # ignore initial double dash arguments though, which are
    # global
    i = 1
    while i < len(sys.argv) and sys.argv[i].startswith("--"):
        i += 1

    # now we get to the command.  strip a single / or - from it.
    if len(sys.argv) > i:
        # turn slash option into a command, stripping the slash
        if sys.argv[i].startswith("/"):
            sys.argv[i] = sys.argv[i][1:]
        elif sys.argv[i].startswith("-") and not sys.argv[i].startswith("--"):
            sys.argv[i] = sys.argv[i][1:]

    # also allow single dash options, by turning them into two dash options
    for i, arg in enumerate(sys.argv[1:]):
        if arg.startswith("-") and not arg.startswith("--"):
            sys.argv[i + 1] = f"--{arg[1:]}"

    tools.click_main(cli, obj={})


if __name__ == "__main__":
    main()



from __future__ import annotations

import contextlib
import ntpath
import winreg
from typing import Any, Iterator, Optional, Tuple, Union
from winreg import HKEYType

"""helper classes for registry operations"""

# map the winreg root key constants to their names
root_keys: dict[int, str] = {}
for key, val in winreg.__dict__.items():
    if key.startswith("HKEY_"):
        root_keys[val] = key


def handle_to_str(handle: int) -> str:
    """Converts a handle to a string"""
    try:
        return root_keys[handle]
    except KeyError:
        return repr(handle)


def join_names(*names: str) -> str:
    """Joins the names together, removing any empty strings"""
    return ntpath.join(*names)


class Key:
    """A key in the registry.  This is a wrapper around the winreg module."""

    @classmethod
    def _create_root_key(cls, root: int, *args, **kwargs):
        """Creates a key object for a root key"""
        if args:
            return cls(root, *args, **kwargs)
        key = cls(root, "", **kwargs)
        key._handle = root
        key.name = ""
        assert key.is_root()
        return key

    @classmethod
    def classes_root(cls, *args, **kwargs):
        """Returns a key object for HKEY_CLASSES_ROOT"""
        return cls._create_root_key(winreg.HKEY_CLASSES_ROOT, *args, **kwargs)

    @classmethod
    def current_user(cls, *args, **kwargs):
        """Returns a key object for HKEY_CURRENT_USER"""
        return cls._create_root_key(winreg.HKEY_CURRENT_USER, *args, **kwargs)

    @classmethod
    def local_machine(cls, *args, **kwargs):
        """Returns a key object for HKEY_LOCAL_MACHINE"""
        return cls._create_root_key(winreg.HKEY_LOCAL_MACHINE, *args, **kwargs)

    @classmethod
    def users(cls, *args, **kwargs):
        """Returns a key object for HKEY_USERS"""
        return cls._create_root_key(winreg.HKEY_USERS, *args, **kwargs)

    @classmethod
    def current_config(cls, *args, **kwargs):
        """Returns a key object for HKEY_CURRENT_CONFIG"""
        return

    def __init__(
        self,
        parent: Union[Key, int],
        *names: str,
        **kwargs,
    ) -> None:
        """Create a new key object.  The key is not opened or created."""
        self._parent = parent
        assert names
        self.name = ntpath.join(*names)
        self._handle: Optional[Union[HKEYType, int]] = None
        self._kwargs = kwargs  # passed to an implicit open() in context manager

    def _hkey_name(self) -> Tuple[Any, str]:
        """returns a handle and name for the key.  Used internally."""
        if isinstance(self._parent, Key):
            if self._parent.is_open():
                return self._parent._handle, self.name
            h, n = self._parent._hkey_name()
            return h, join_names(n, self.name)
        return self._parent, self.name

    def _hkey_fullname(self) -> Tuple[Any, str]:
        """returns a top level handle and full name for the key.  Used internally."""
        if isinstance(self._parent, Key):
            h, n = self._parent._hkey_fullname()
            return h, join_names(n, self.name)
        return self._parent, self.name

    def __repr__(self) -> str:
        """Returns a string representation of the key"""
        h, n = self._hkey_fullname()
        return f"Key<{handle_to_str(h)}:{n!r}>"

    def open(
        self,
        *subkeys: str,
        create: bool = False,
        write: Optional[bool] = None,
        check: bool = True,
    ) -> Optional[Key]:
        """Opens an existing key"""
        if subkeys:
            return self.subkey(*subkeys).open(create=create, write=write, check=check)
        assert self._handle is None
        handle, name = self._hkey_name()
        try:
            # create defaults for true if create is true, false otherwise
            write = write if write is not None else create
            access: int = winreg.KEY_READ
            if write:
                access |= winreg.KEY_WRITE
            func = winreg.CreateKeyEx if create else winreg.OpenKeyEx
            self._handle = func(handle, name, access=access)
        except FileNotFoundError as e:
            if check:
                raise ValueError(f"Key {self.name!r} not found") from e
            return None
        return self

    def subkey(self, *subkeys: str, **kwargs) -> Key:
        """Returns a subkey object"""
        return Key(self, *subkeys, **kwargs)

    def __call__(self, *subkeys: str, **kwargs) -> Key:
        """Returns a subkey object"""
        return self.subkey(*subkeys, **kwargs)

    def exists(self) -> bool:
        """checks if the key exists"""
        if self.is_open():
            return True
        if self.open(check=False) is None:
            return False
        self.close()
        return True

    def __bool__(self) -> bool:
        """checks if the key exists"""
        return self.exists()

    def is_open(self) -> bool:
        """Checks if the key is opened"""
        return self._handle is not None

    def is_root(self) -> bool:
        """Checks if the key is a root key"""
        return self._parent == self._handle

    def create(self, *args, **kwargs) -> Optional[Key]:
        """alias for open(... create=True)"""
        kwargs["create"] = True
        return self.open(*args, **kwargs)

    def close(self) -> None:
        """Closes the key."""
        # close if open and if not a root key
        if self.is_open() and not self.is_root():
            handle, self._handle = self._handle, None
            assert handle is not None
            winreg.CloseKey(handle)

    # iterating over the key/value pairs (items) in the key, similar to a dict.

    def items(self) -> Iterator[tuple[str, tuple[Any, int]]]:
        """Iterates over the values in the key"""
        assert self._handle is not None
        for i in range(1000):
            try:
                name, value, type = winreg.EnumValue(self._handle, i)
                yield (name, (value, type))
            except OSError:
                break

    def keys(self) -> Iterator[str]:
        """iterates of the item names in the key"""
        for itemname, _ in self.items():
            yield itemname

    def values(self) -> Iterator[tuple[Any, int]]:
        """iterates of the item values in the key"""
        for _, value in self.items():
            yield value

    def stringitems(self) -> Iterator[tuple[str, str]]:
        """Iterates over the string items in the key"""
        for k, (v, t) in self.items():
            if t == winreg.REG_SZ:
                yield (k, v)

    def subkeys(self) -> Iterator[Key]:
        """Iterates over the subkeys in the key"""
        assert self._handle is not None
        for i in range(1000):
            try:
                name = winreg.EnumKey(self._handle, i)
                yield Key(self, name)
            except OSError:
                break

    def value_get(self, name: str, default: Any = None) -> tuple[Any, int]:
        """Gets a value from the key, returning the value and a type."""
        assert self._handle is not None
        try:
            return winreg.QueryValueEx(self._handle, name)
        except FileNotFoundError:
            raise KeyError(name)

    def value_set(self, name: str, value: Any, type=winreg.REG_SZ) -> None:
        """Sets a value in the key"""
        assert self._handle is not None
        winreg.SetValueEx(self._handle, name, 0, type, value)

    def value_del(self, name: str) -> None:
        assert self._handle is not None
        winreg.DeleteValue(self._handle, name)

    def get(self, name: str, default: Any = None) -> Any:
        """Gets a value from the key"""
        try:
            return self[name]
        except KeyError:
            return default

    def __getitem__(self, name: str) -> Tuple[Any, int]:
        """Get a value from the key"""
        assert self._handle is not None
        try:
            v, t = winreg.QueryValueEx(self._handle, name)
            if t == winreg.REG_BINARY and v is None:
                v = b""
            return v
        except FileNotFoundError as e:
            raise KeyError(name) from e

    def __setitem__(self, name: str, value: Any) -> None:
        """Sets a value in the key. We assume a string"""
        if isinstance(value, tuple):
            return self.value_set(name, *value)
        elif isinstance(value, int):
            return self.value_set(name, value, winreg.REG_DWORD)
        elif isinstance(value, bytes):
            return self.value_set(name, value, winreg.REG_BINARY)
        elif value is None:
            return self.value_set(name, None, winreg.REG_NONE)
        # default handling
        return self.value_set(name, value, winreg.REG_SZ)

    def __delitem__(self, name: str) -> None:
        """Deletes a value from the key"""
        assert self._handle is not None
        try:
            winreg.DeleteValue(self._handle, name)
        except FileNotFoundError as e:
            raise KeyError(name) from e

    def __enter__(self) -> Optional[Key]:
        """Context manager: opens the key if required"""
        if not self.is_open():
            return self.open(**self._kwargs)
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    @contextlib.contextmanager
    def opened(self, *args, **kwargs):
        """Context manager: opens the key if required"""
        if args or not self.is_open():
            with self.open(*args, **kwargs) as key:
                yield key
        else:
            yield self

    def delete(self, tree: bool = False, missing_ok=True) -> None:
        """Deletes the key, optionally recursively."""
        if self.is_open():
            raise ValueError("Cannot delete open key")
        if missing_ok and not self.exists():
            return
        if tree:
            with self.opened():
                for subkey in list(self.subkeys()):
                    subkey.delete(tree=True)
        h, n = self._hkey_name()
        winreg.DeleteKey(h, n)

    def copy(self) -> Key:
        """Returns a fresh, un-opened copy of the key"""
        return Key(self._parent, self.name)

    def print(self, tree=False, indent=4, level=0):
        """Prints the key to stdout"""
        print(" " * level * indent + f"key: '{self.name}'")
        with self.opened() as key:
            for name, value in key.items():
                print(" " * (level + 1) * indent + f"val: '{name}' = {value})")
            if not tree:
                for sub in key.subkeys():
                    print(" " * (level + 1) * indent + f"key: '{sub.name}'")
            else:
                for sub in key.subkeys():
                    sub.print(tree=True, indent=indent, level=level + 1)

    def as_dict(self):
        """Returns the key and subkeys as a dictionary"""
        with self.opened() as key:
            return {
                "keys": {sub.name: sub.as_dict() for sub in key.subkeys()},
                "values": {name: value for name, (value, _) in key.items()},
            }

    def from_dict(self, data, remove: bool = False):
        """Sets the key and subkeys from a dictionary"""
        with self.opened(create=True) as key:
            for name, value in data["values"].items():
                if isinstance(value, tuple):
                    key.set_value(name, *value)
                else:
                    key[name] = value
            for subname, subdata in data["keys"].items():
                key.create(subname).from_dict(subdata, remove=remove)

            if remove:
                for sub in key.subkeys():
                    if sub.name not in data["keys"]:
                        sub.delete(tree=True)
                for name in key.keys():
                    if name not in data["values"]:
                        del key[name]



from __future__ import annotations

import glob
import os.path
from configparser import SectionProxy
from dataclasses import dataclass, field

from . import config_cache, paths

# dummy implementation of the platform module, we only ever use the empty platform.

has_searched_for_files = False
has_searched_for_platforms = False

data_driven_ini_filenames: list[str] = []
platforms: dict[str, DataDrivenPlatformInfo] = {}
platform_name_aliases: dict[str, str] = {}
all_sorted_platforms: list[str] = []
sorted_platforms: list[str] = []
all_sorted_platform_infos: list[DataDrivenPlatformInfo] = []
sorted_platform_infos: list[DataDrivenPlatformInfo] = []


def reset():
    global has_searched_for_files
    global has_searched_for_platforms
    has_searched_for_files = False
    has_searched_for_platforms = False


@dataclass
class DataDrivenPlatformInfo:
    ini_platform_name: str = ""
    ini_parent_chain: list[str] = field(default_factory=list)
    is_fake_platform: bool = False
    is_confidential: bool = False


def get_info(platform: str) -> DataDrivenPlatformInfo:
    if platform:
        return get_platforms().get(platform, DataDrivenPlatformInfo())
    return DataDrivenPlatformInfo()


# hardwired for windows now
def ini_platform_name() -> str:
    return "Windows"


def platform_name() -> str:
    return "Windows"


def get_data_driven_ini_filenames() -> list[str]:
    global has_searched_for_files
    global data_driven_ini_filenames
    if has_searched_for_files:
        return data_driven_ini_filenames

    # look for the special files in any config subdirectories
    patterns = [
        # look for the special files in any config subdirectories
        os.path.join(paths.engine_config_dir(), "**", "DataDrivenPlatformInfo.ini"),
        os.path.join(
            paths.engine_patform_extensions_dir(), "**", "DataDrivenPlatformInfo.ini"
        ),
        os.path.join(
            paths.project_patform_extensions_dir(), "**", "DataDrivenPlatformInfo.ini"
        ),
    ]
    found = []
    for p in patterns:
        found += glob.glob(p, recursive=True)
    data_driven_ini_filenames = found
    has_searched_for_files = True
    return data_driven_ini_filenames


def get_platforms() -> dict[str, DataDrivenPlatformInfo]:
    global has_searched_for_platforms
    global platforms
    if has_searched_for_platforms:
        return platforms

    # get the filenames
    filenames = get_data_driven_ini_filenames()

    # read the files
    platforms = {}
    ini_parents: dict[str, str] = {}
    for filename in filenames:
        platformstring, inifile = load_data_driven_ini_file(filename)

        if "DataDrivenPlatformInfo" not in inifile.sections():
            continue

        if platformstring not in platforms:
            platforms[platformstring] = DataDrivenPlatformInfo()
        info = platforms[platformstring]
        ddpi = inifile["DataDrivenPlatformInfo"]
        load_ddpi_ini_settings(platformstring, info, ddpi)
        info.ini_platform_name = platformstring

        ini_parent = ddpi.get("ParentPlatform", "")
        ini_parents[platformstring] = ini_parent

        aliases = ddpi.get("PlatformNameAliases", "")
        for alias in aliases.split(","):
            alias = alias.strip()
            if alias:
                platform_name_aliases[alias] = platformstring

    # now that all are read in, calculate the ini parent chain, starting with parent-most
    for platform, info in platforms.items():
        # walk up the chain and build up the ini chain of parents
        current_platform = ini_parents.get(platform, "")
        while current_platform:
            info.ini_parent_chain.append(current_platform)
            current_platform = ini_parents.get(current_platform, "")
        info.ini_parent_chain.reverse()

    global all_sorted_platforms
    global sorted_platforms
    global all_sorted_platform_infos
    global sorted_platform_infos
    all_sorted_platforms = list(platforms.keys())
    all_sorted_platforms.sort()
    sorted_platforms = [
        x for x in all_sorted_platforms if not platforms[x].is_fake_platform
    ]
    all_sorted_platform_infos = [platforms[x] for x in all_sorted_platforms]
    sorted_platform_infos = [platforms[x] for x in sorted_platforms]

    has_searched_for_platforms = True
    return platforms


def load_data_driven_ini_file(filename) -> tuple[str, config_cache.UEConfigParser]:
    parser = config_cache.UEConfigParser()
    parser.read(filename)

    if filename.startswith(
        paths.engine_patform_extensions_dir()
    ) or filename.startswith(paths.project_patform_extensions_dir()):
        # platforms/platform/config/DataDrivenPlatformInfo.ini
        platform_name = os.path.basename(os.path.dirname(os.path.dirname(filename)))
    else:
        # config/platform/DatDrivenPlatformInfo.ini
        platform_name = os.path.basename(os.path.dirname(filename))
    return platform_name, parser


def load_ddpi_ini_settings(
    platform_name: str, info: DataDrivenPlatformInfo, ini: SectionProxy
) -> None:
    # don't support command line prefixes

    def get_bool(name, default):
        return ini.getboolean(name, default)

    info.is_confidential = get_bool("bIsConfidential", info.is_confidential)
    info.is_fake_platform = get_bool("bIsFakePlatform", info.is_fake_platform)

    """

	DDPIGetString(IniFile, TEXT("TargetSettingsIniSectionName"), Info.TargetSettingsIniSectionName);
	DDPIGetString(IniFile, TEXT("HardwareCompressionFormat"), Info.HardwareCompressionFormat);
	DDPIGetStringArray(IniFile, TEXT("AdditionalRestrictedFolders"), Info.AdditionalRestrictedFolders);

	DDPIGetBool(IniFile, TEXT("Freezing_b32Bit"), Info.Freezing_b32Bit);
	DDPIGetUInt(IniFile, Info.Freezing_b32Bit ? TEXT("Freezing_MaxFieldAlignment32") : TEXT("Freezing_MaxFieldAlignment64"), Info.Freezing_MaxFieldAlignment);
	DDPIGetBool(IniFile, TEXT("Freezing_bForce64BitMemoryImagePointers"), Info.Freezing_bForce64BitMemoryImagePointers);
	DDPIGetBool(IniFile, TEXT("Freezing_bAlignBases"), Info.Freezing_bAlignBases);

	DDPIGetGuid(IniFile, TEXT("GlobalIdentifier"), Info.GlobalIdentifier);
	checkf(Info.GlobalIdentifier != FGuid(), TEXT("Platform %s didn't have a valid GlobalIdentifier set in DataDrivenPlatformInfo.ini"), *PlatformName.ToString());

	// NOTE: add more settings here!
	DDPIGetBool(IniFile, TEXT("bHasDedicatedGamepad"), Info.bHasDedicatedGamepad);
	DDPIGetBool(IniFile, TEXT("bDefaultInputStandardKeyboard"), Info.bDefaultInputStandardKeyboard);

	DDPIGetBool(IniFile, TEXT("bInputSupportConfigurable"), Info.bInputSupportConfigurable);
	DDPIGetString(IniFile, TEXT("DefaultInputType"), Info.DefaultInputType);
	DDPIGetBool(IniFile, TEXT("bSupportsMouseAndKeyboard"), Info.bSupportsMouseAndKeyboard);
	DDPIGetBool(IniFile, TEXT("bSupportsGamepad"), Info.bSupportsGamepad);
	DDPIGetBool(IniFile, TEXT("bCanChangeGamepadType"), Info.bCanChangeGamepadType);
	DDPIGetBool(IniFile, TEXT("bSupportsTouch"), Info.bSupportsTouch);

#if DDPI_HAS_EXTENDED_PLATFORMINFO_DATA

	DDPIGetString(IniFile, TEXT("AutoSDKPath"), Info.AutoSDKPath);
	DDPIGetString(IniFile, TEXT("TutorialPath"), Info.SDKTutorial);
	DDPIGetName(IniFile, TEXT("PlatformGroupName"), Info.PlatformGroupName);
	DDPIGetName(IniFile, TEXT("PlatformSubMenu"), Info.PlatformSubMenu);


	DDPIGetString(IniFile, TEXT("NormalIconPath"), Info.IconPaths.NormalPath);
	DDPIGetString(IniFile, TEXT("LargeIconPath"), Info.IconPaths.LargePath);
	DDPIGetString(IniFile, TEXT("XLargeIconPath"), Info.IconPaths.XLargePath);
	if (Info.IconPaths.XLargePath == TEXT(""))
	{
		Info.IconPaths.XLargePath = Info.IconPaths.LargePath;
	}

	FString PlatformString = PlatformName.ToString();
	Info.IconPaths.NormalStyleName = *FString::Printf(TEXT("Launcher.Platform_%s"), *PlatformString);
	Info.IconPaths.LargeStyleName = *FString::Printf(TEXT("Launcher.Platform_%s.Large"), *PlatformString);
	Info.IconPaths.XLargeStyleName = *FString::Printf(TEXT("Launcher.Platform_%s.XLarge"), *PlatformString);

	Info.bCanUseCrashReporter = true; // not specified means true, not false
	DDPIGetBool(IniFile, TEXT("bCanUseCrashReporter"), Info.bCanUseCrashReporter);
	DDPIGetBool(IniFile, TEXT("bUsesHostCompiler"), Info.bUsesHostCompiler);
	DDPIGetBool(IniFile, TEXT("bUATClosesAfterLaunch"), Info.bUATClosesAfterLaunch);
	DDPIGetBool(IniFile, TEXT("bIsEnabled"), Info.bEnabledForUse);

	DDPIGetName(IniFile, TEXT("UBTPlatformName"), Info.UBTPlatformName);
	// if unspecified, use the ini platform name (only Win64 breaks this)
	if (Info.UBTPlatformName == NAME_None)
	{
		Info.UBTPlatformName = PlatformName;
	}
	Info.UBTPlatformString = Info.UBTPlatformName.ToString();
		
	GCommandLinePrefix = TEXT("");

	// now that we have all targetplatforms in a single TP module per platform, just look for it (or a ShaderFormat for other tools that may want this)
	// we could look for Platform*, but then platforms that are a substring of another one could return a false positive (Windows* would find Windows31TargetPlatform)
	Info.bHasCompiledTargetSupport = FDataDrivenPlatformInfoRegistry::HasCompiledSupportForPlatform(PlatformName, FDataDrivenPlatformInfoRegistry::EPlatformNameType::TargetPlatform);

	ParsePreviewPlatforms(IniFile);
#endif
}
"""



# stuff to install visual studio
import os.path
import subprocess

import click
import requests
from click import echo

bootstrapper = "https://aka.ms/vs/17/release/vs_community.exe"

install_args = """\
--add Microsoft.VisualStudio.Workload.NativeGame --includeRecommended
--add Microsoft.Net.Component.4.6.2.TargetingPack
--add Microsoft.NetCore.Component.Runtime.6.0
--add Microsoft.Net.Component.4.8.SDK
--add Microsoft.Net.ComponentGroup.4.8.DeveloperTools
--add Microsoft.Component.MSBuild
--add Microsoft.VisualStudio.Component.NuGet
--add Microsoft.NetCore.Component.SDK
--wait --passive --norestart
"""
install_args = " ".join(install_args.split())

update_args = """\
modify --channelid VisualStudio.17.Release --productid Microsoft.VisualStudio.Product.Community
--add Microsoft.VisualStudio.Workload.NativeGame --includeRecommended
--add Microsoft.Net.Component.4.6.2.TargetingPack
--add Microsoft.NetCore.Component.Runtime.6.0
--add Microsoft.Net.Component.4.8.SDK
--add Microsoft.Net.ComponentGroup.4.8.DeveloperTools
--add Microsoft.Component.MSBuild
--add Microsoft.VisualStudio.Component.NuGet
--add Microsoft.NetCore.Component.SDK
--wait --passive --norestart
"""
update_args = " ".join(update_args.split())


def ensure_bootstrapper():
    """Check if bootstrapper exists in tmp folder, otherwise, download it to there."""

    # get temp directory
    tmp = os.getenv("TEMP")
    tmpname = os.path.join(tmp, os.path.basename(bootstrapper))
    if os.path.isfile(tmpname):
        return tmpname

    # download bootstrapper using requests library
    print(f"Downloading bootstrapper to {tmpname}")
    r = requests.get(bootstrapper)
    r.raise_for_status()
    with open(tmpname, "wb") as f:
        f.write(r.content)

    return tmpname


def install_vs():
    """Install visual studio"""

    # ensure bootstrapper exists
    bootstrapper = ensure_bootstrapper()

    # run bootstrapper
    subprocess.run(f"{bootstrapper} {install_args}", shell=True, check=True)
    echo("Visual studio installed.")


def update_vs():
    """Install visual studio"""

    # ensure bootstrapper exists
    bootstrapper = ensure_bootstrapper()

    # run bootstrapper
    r = subprocess.run(f"{bootstrapper} {update_args}", shell=True)
    # it will fail with exit status 1 if it is already up to date
    if r.returncode not in (0, 1):
        r.check_returncode()

    if r.returncode == 1:
        echo("Visual studio is already up to date.")
    else:
        echo("Visual studio updated.")


@click.group()
def vstudio():
    """Manage visual studio."""
    pass


@vstudio.command()
def install():
    """Install visual studio."""
    install_vs()


@vstudio.command()
def update():
    """Update visual studio."""
    update_vs()



import ctypes
import ctypes.wintypes
import os.path
import shlex
import uuid
from typing import Dict, Optional

import click

from ..registry import Key
from . import desktop


def enumerate_engine_installations() -> Dict[str, str]:
    launcher_engines = enumerate_launcher_engine_installations()
    user_engines = enumerate_user_engine_installations()
    launcher_engines.update(user_engines)
    return launcher_engines


def enumerate_launcher_engine_installations() -> Dict[str, str]:
    """Enumerate the engine installations from the launcher"""

    result: Dict[str, str] = {}
    with Key.local_machine(
        "SOFTWARE\\EpicGames\\Unreal Engine",
        check=False,
    ) as key:
        if not key:
            return result
        for subkey in key.subkeys():
            with subkey:
                name = subkey.name
                root = subkey.get("InstalledDirectory", None)
                if root and desktop.is_valid_root_directory(root):
                    result[name] = desktop.normalize_root(root)
    return result


def enumerate_user_engine_installations() -> Dict[str, str]:
    """Enumerate the engine installations from the launcher"""

    result: Dict[str, str] = {}
    remove = []  # values to remove
    unique_directories = set()

    with Key.current_user(
        "SOFTWARE\\Epic Games\\Unreal Engine\\Builds", write=True
    ) as key:
        if not key:
            return result
        for k, (v, _) in key.items():
            n = desktop.normalize_root(v)
            ok = False
            if desktop.is_valid_root_directory(v):
                try:
                    guid = uuid.UUID(k)
                except ValueError:
                    guid = None
                if guid:
                    # clean out duplicate guid keys for the same engine
                    if n not in unique_directories:
                        unique_directories.add(n)
                        ok = True
                else:
                    # we accept all non-guid keys, since they were created by the user
                    ok = True
            if ok:
                result[k] = v
            else:
                remove.append(k)

        # Remove all the keys which weren't valid
        for k in remove:
            del key[k]
    return result


def register_engine_installation(
    root_dir: str, engine_id: Optional[str] = None
) -> Optional[str]:
    """Register an engine installation"""
    if not desktop.is_valid_root_directory(root_dir):
        return None
    if not engine_id:
        engine_id = str(uuid.uuid4())
    return register_engine_installation_identifier(root_dir, engine_id)


def register_engine_installation_identifier(root_dir: str, engine_id: str) -> str:
    with Key.current_user(
        "SOFTWARE\\Epic Games\\Unreal Engine\\Builds", create=True
    ) as key:
        if desktop.is_guid(engine_id):
            root_dir = desktop.normalize_root(root_dir)
            # check if there is a guid key with the same root dir
            for k, (v, _) in key.items():
                if desktop.is_guid(k) and desktop.normalize_root(v) == root_dir:
                    # we just re-use the existing guid key
                    return k
        key[engine_id] = root_dir
    return engine_id


def update_file_associations(
    engine_root=None, handler=None, user=True, verify=True, test=False
):
    # the engine contains a special execution to act as a shell handler
    # for files with the .uproject extension

    if not handler:
        handler = os.path.join(
            engine_root, "Engine", "Binaries", "Win64", "UnrealVersionSelector.exe"
        )
        handler = [os.path.abspath(handler)]
    elif not isinstance(handler, list):
        handler = [handler]
    handler[0] = os.path.abspath(handler[0])

    # we need to find an icon for the handler
    paxoicon = os.path.abspath(
        os.path.join(os.path.dirname(__file__), "..", "paxo.ico")
    )
    iconhandler = handler[0]
    if not os.path.exists(handler[0]):
        iconhandler = paxoicon
    # or, if it is not UnrealVersionSelector, use the icon from paxo
    if "UnrealVersion" not in os.path.basename(handler[0]):
        iconhandler = paxoicon

    # quoted handler for the registry
    quoted_handler = " ".join(f'"{f}"' for f in handler)
    quoted_iconhandler = f'"{iconhandler}"'
    quoted_paxoicon = f'"{paxoicon}"'
    set_associations(user, quoted_handler, quoted_iconhandler, quoted_paxoicon)


def set_associations(user: bool, handler: str, fileicon: str, cmdicon: str):
    # we must now find the appropriate place in the registry and add it.abs
    # this is a bit tricky, but we can use the python winreg module
    if user:
        root = Key.current_user("Software\\Classes")
    else:
        root = Key.local_machine("Software\\Classes")
    with root.create(".uproject") as key:
        key[""] = "Unreal.ProjectFile"
        print(key[""])

    with root.create("Unreal.ProjectFile") as key:
        # we could well clear this subtree first and write it all fresh
        key[""] = "Unreal Engine Project File"
        print(key[""])
        key["VersionSelectorExecutable"] = handler

        # the DefaultIcon subkey, used for the shell in the file
        with key.create("DefaultIcon") as subkey:
            subkey[""] = fileicon

        with key.create("shell") as subkey:

            def add(name, description, command):
                with subkey.create(name) as k2:
                    k2[""] = description
                    with k2.create("command") as k3:
                        k3[""] = command
                    with k2.create("icon") as k3:
                        k3[""] = cmdicon

            add("open", "Open", handler + ' /editor "%1"')
            add("run", "Launch game", handler + ' /game "%1"')
            add(
                "rungenproj",
                "Generate Visual Studio project files",
                handler + ' /projectfiles "%1"',
            )
            add(
                "switchversion",
                "Switch Unreal Engine version...",
                handler + ' /switchversion "%1"',
            )
            add("hotdog", "Get some sausages...", "explorer.exe https://hot-dog.org")

    # If the user has manually selected something other than our extension,
    # we need to delete it.
    # Explorer explicitly disables set access on that keys
    # in that folder, but we can delete the whole thing.
    user_choice = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\FileExts\\.uproject\\UserChoice"
    key = Key.current_user(user_choice)
    if key:
        key.delete(tree=True)
    return handler


def verify_file_associations(user=True, silent=False):
    try:
        if user:
            root = Key.current_user("Software\\Classes")
        else:
            root = Key.local_machine("Software\\Classes")

        with root(".uproject", check=False) as key:
            if not key or key[""] != "Unreal.ProjectFile":
                raise click.ClickException(
                    f"File association for .uproject not set for scope {'user' if user else 'machine'}"
                )
        with root("Unreal.ProjectFile", check=False) as key:
            if not key:
                raise click.ClickException(
                    f"File association for Unreal.ProjectFile not set for scope {'user' if user else 'machine'}"
                )

            with key("shell", "open", "command", check=False) as key:
                if not key:
                    raise click.ClickException(
                        f"Unreal.ProjectFile shell not set for scope {'user' if user else 'machine'}"
                    )

                try:
                    handler = key[""]
                except KeyError:
                    handler = ""
                if not handler:
                    raise click.ClickException(
                        f"Unreal.ProjectFile shell open command not set for scope {'user' if user else 'machine'}"
                    )
                parts = shlex.split(handler)
                try:
                    editor_index = parts.index("/editor")
                except ValueError:
                    raise click.ClickException(
                        f"Unreal.ProjectFile shell open command does not contain /editor for scope {'user' if user else 'machine'}"
                    )
                return parts[:editor_index]
    except click.ClickException:
        if silent:
            return False
        raise


def clear_file_associations(user=True):
    if user:
        root = Key.current_user("Software\\Classes")
    else:
        root = Key.local_machine("Software\\Classes")

    key = root(".uproject")
    if key:
        key.delete(tree=True)
    key = root("Unreal.ProjectFile")
    if key:
        key.delete(tree=True)


def try_read_msbuild_install_path() -> Optional[str]:
    # first try the location in program files
    program_files = get_program_files_folder()
    if program_files:
        tool_path = os.path.join(program_files, "MSBuild/14.0/bin/MSBuild.exe")
        if os.path.exists(tool_path):
            return tool_path

    # Try to get the MSBuild 14.0 path directly (see https://msdn.microsoft.com/en-us/library/hh162058(v=vs.120).aspx)
    path = try_lookup_msbuild_install_path(
        "Microsoft\\MSBuild\\ToolsVersions\\14.0", "MSBuildToolsPath", "MSBuild.exe"
    )

    # Check for MSBuild 15. This is installed alongside Visual Studio 2017, so we get the path relative to that.
    path = path or try_lookup_msbuild_install_path(
        "Microsoft\\VisualStudio\\SxS\\VS7", "15.0", "MSBuild\\15.0\\bin\\MSBuild.exe"
    )

    # Check for older versions of MSBuild. These are registered as separate versions in the registry.
    path = path or try_lookup_msbuild_install_path(
        "Microsoft\\MSBuild\\ToolsVersions\\12.0", "MSBuildToolsPath", "MSBuild.exe"
    )
    path = path or try_lookup_msbuild_install_path(
        "Microsoft\\MSBuild\\ToolsVersions\\4.0", "MSBuildToolsPath", "MSBuild.exe"
    )
    return path


def get_known_folder_path(folder_id: uuid.UUID) -> str:
    dll = ctypes.windll.shell32
    buf = ctypes.c_wchar_p()
    dll.SHGetKnownFolderPath(
        folder_id.bytes_le, ctypes.wintypes.DWORD(), None, ctypes.byref(buf)
    )
    result = buf.value
    ctypes.windll.Ole32.CoTaskMemFree(buf)
    return result or ""


def get_program_files_folder() -> str:
    """Get the program files folder using the proper ShGetKnowFolderPath function.  This is the correct way to do it."""
    FOLDERID_ProgramFiles = uuid.UUID("{905e63b6-c1bf-494e-b29c-65b732d3d21a}")
    return get_known_folder_path(FOLDERID_ProgramFiles)


_user_settings_dir = ""


def user_settings_dir() -> str:
    global _user_settings_dir
    if not _user_settings_dir:
        FOLDERID_LocalAppData = uuid.UUID("{F1B32785-6FBA-4FCF-9D55-7B8E7F157091}")
        _user_settings_dir = get_known_folder_path(FOLDERID_LocalAppData)
    return _user_settings_dir


_user_dir = ""


def user_dir() -> str:
    global _user_dir
    if not _user_dir:
        FOLDERID_Documents = uuid.UUID("{FDD39AD0-238F-46AF-ADB4-6C85480369C7}")
        _user_dir = get_known_folder_path(FOLDERID_Documents)
    return _user_dir


def try_lookup_msbuild_install_path(
    key_relative: str, value_name: str, relative_path: str
) -> Optional[str]:
    def lookup(rootkey: Key, key_name: str) -> Optional[str]:
        with rootkey(key_name, key_relative, check=False) as key:
            if key is not None:
                value = key.get(value_name, None)
                if value:
                    fn = os.path.join(value, relative_path)
                    if os.path.exists(fn):
                        return fn
        return None

    path = lookup(Key.current_user(), "SOFTWARE")
    path = path or lookup(Key.local_machine(), "SOFTWARE")
    path = path or lookup(Key.current_user(), "SOFTWARE\\Wow6432Node")
    path = path or lookup(Key.local_machine(), "SOFTWARE\\Wow6432Node")
    return path


def run_unreal_build_tool(msg, root_dir, args, warn) -> bool:
    # Get the path to UBT
    unreal_build_tool_path = desktop.get_unreal_build_tool_executable_filename(root_dir)
    if not unreal_build_tool_path or not os.path.isfile(unreal_build_tool_path):
        warn.log(
            f"Couldn't find UnrealBuildTool at '{unreal_build_tool_path}'", type="ERROR"
        )
        return False

    warn.log(f"Running {unreal_build_tool_path} {args}")

    success, returncode = warn.run_process(msg, [unreal_build_tool_path] + args)
    return success and returncode == 0



import glob
import json
import os.path
import subprocess
import sys
import winreg
from contextlib import contextmanager

import click
from click import echo, secho

from . import p4, tools, uvs
from .registry import Key
from .uebase import desktop
from .uebase.desktop import is_valid_root_directory


# helper to close winreg keys
@contextmanager
def close_key(key):
    try:
        yield key
    finally:
        key.Close()


# various tools for setting up UE environment and running it.abs


@click.group()
@click.pass_context
def ue(ctx):
    """work with Unreal Engine"""


@ue.group()
@click.option(
    "--engine-path",
    type=click.Path(exists=False),
    default=lambda: p4.get_engine_path(),
    help="The path to register if not default engine path.",
)
@click.option(
    "--scope",
    type=click.Choice(["user", "machine"]),
    default="user",
    help="Work with user or machine registry",
)
@click.option(
    "--check/--no-check", is_flag=True, help="Only accept valid engine paths."
)
@click.pass_context
def engine(ctx, engine_path, scope, check):
    """work with engine registration"""
    ctx.ensure_object(dict)
    ctx.obj["engine_path"] = engine_path
    ctx.obj["scope"] = scope
    ctx.obj["reg_root"] = (
        winreg.HKEY_CURRENT_USER if scope == "user" else winreg.HKEY_LOCAL_MACHINE
    )
    ctx.obj["check"] = check


@ue.command()
def setup():
    setup_ue(tools.workdrive_get(), p4.get_engine_path(), tools.location_get())


@engine.command("register")
@click.pass_context
@click.option("--engine-id", type=str, metavar="<id-string>", default=None)
def engine_register(ctx, engine_id):
    check = ctx.obj["check"]
    register_engine(ctx.obj["engine_path"], engine_id, check=check)


@engine.command("deregister")
@click.argument("engine-id", type=str, metavar="<id-string>")
@click.pass_context
def engine_deregister(ctx, engine_id):
    deregister_engine(ctx.obj["engine_path"], engine_id)


@engine.command("list")
@click.pass_context
@click.option("--all", is_flag=True, help="List both user and machine registrations.")
def engine_list(ctx, all):
    user = ctx.obj["scope"] == "user"
    if user or all:
        print("User registered engines:")
        reg_root = winreg.HKEY_CURRENT_USER
        with Key(reg_root, "SOFTWARE\\Epic Games\\Unreal Engine\\Builds") as key:
            for name, value in key.items():
                print(f"{name} = {value[0]}")
    if not user or all:
        print("System registered engines:")
        reg_root = winreg.HKEY_LOCAL_MACHINE
        key = Key(reg_root, "SOFTWARE\\EpicGames\\Unreal Engine")
        if key:
            with key:
                for subkey in key.subkeys():
                    with subkey:
                        print(f"{subkey.name} = {subkey['InstalledDirectory']}")


@engine.command("lookup")
@click.pass_context
@click.argument("engine-id", type=str, metavar="<id-string>")
@click.option(
    "--all/--no-all",
    is_flag=True,
    default=True,
    help="Look up in both user and machine registrations.",
)
def engine_lookup(ctx, engine_id, all, check):
    user = ctx.obj["scope"] == "user"
    path = None
    if user or all:
        reg_root = winreg.HKEY_CURRENT_USER
        with Key(reg_root, "SOFTWARE\\Epic Games\\Unreal Engine\\Builds") as key:
            if key:
                path = key.get(engine_id, None)
    if path is not None and check and not is_valid_root_directory(path):
        path = None
    if path is None and (not user or all):
        reg_root = winreg.HKEY_LOCAL_MACHINE
        with Key(reg_root, "SOFTWARE\\EpicGames\\Unreal Engine") as key:
            if key:
                for subkey in key.subkeys():
                    with subkey:
                        if subkey.name == engine_id:
                            path = subkey["InstalledDirectory"]
                            break
    if path is not None and check and not is_valid_root_directory(path):
        path = None

    if path is not None:
        echo(path)
    sys.exit(1)


@ue.group()
@click.option("--scope", type=click.Choice(["user", "machine"]), default="user")
@click.pass_context
def uproject(ctx, scope):
    """work with .uproject files registration"""
    ctx.ensure_object(dict)
    ctx.obj["scope"] = scope


@uproject.command("set")
@click.pass_context
@click.option(
    "--verify/--no-verify", default=True, help="Verify that target files exist."
)
@click.option(
    "--engine-path",
    type=click.Path(exists=False),
    default=lambda: p4.get_engine_path(),
    help="The path to the handling Engine if not the default path.",
)
@click.option(
    "--handler",
    type=click.Path(exists=False),
    default=None,
    help="The path of a custom handler executable to use.",
)
def uproject_register(ctx, verify, engine_path, handler):
    user = ctx.obj["scope"] == "user"
    handler = register_uproject_handler(
        engine_path=engine_path, handler=handler, user=user, verify=verify
    )
    echo(f"set .uproject handler for {ctx.obj['scope']} to {handler!r}")


@uproject.command("get")
@click.pass_context
@click.option("-any/--no-any", default=True, help="Get the handler for any scope.")
@click.option(
    "--verify/--no-verify", default=True, help="Verify that target files exist."
)
def uproject_get(ctx, verify, any):
    scope = ctx.obj["scope"] if not any else "user"
    handler = get_uproject_handler(user=scope == "user", verify=verify)
    if not handler and any:
        scope = "machine"
        handler = get_uproject_handler(user=False, verify=verify)
    if handler:
        echo(f".uproject handler is set to {handler!r} for scope {scope!r} ")
        if not os.path.isfile(handler):
            secho(f"Warning: handler {handler!r} does not exist", fg="red")
    else:
        if not any:
            echo(f".uproject handler is not set for scope {scope!r} ")
        else:
            echo(".uproject handler is not set for any scope")


@uproject.command("show")
@click.pass_context
@click.option(
    "--verify/--no-verify", default=True, help="Verify that target files exist."
)
def uproject_show(ctx, verify):
    user = ctx.obj["scope"] == "user"
    check_uproject_handler(user=user, verify=verify)


@uproject.command("deregister")
@click.pass_context
def uproject_deregister(ctx):
    user = ctx.obj["scope"] == "user"
    deregister_uproject_handler(user=user)


def setup_ue(work_drive, engine_path, location=None):
    # set up the environment variables
    set_ddc_vars(work_drive, location)

    # register the engine
    if not os.path.isfile(os.path.join(engine_path, "build_info.json")):
        raise click.ClickException(
            f"Engine path {engine_path} does not contain an engine.  Have you performed a sync?"
        )

    uvs.update_file_associations(user=True)
    engine_id = get_engine_id_from_build_info(engine_path)
    uvs.register_current_engine_directory(engine_root=engine_path, engine_id=engine_id)
    # register_uproject_handler(engine_path)
    # register_engine(engine_path)

    # install prerequisites
    install_prerequisites(engine_path)


def set_location(location):
    # set the env vars related to location
    loc = tools.locations[location]
    ddc = loc["ddc"]
    if ddc:
        print(f"setting shared data cache path to {ddc} for location {location}")
        tools.env_var_set("UE-SharedDataCachePath", ddc)
    else:
        print(f"clearing shared data cache path for location {location}")
        tools.env_var_del("UE-SharedDataCachePath")


def set_ddc_vars(work_drive, location):
    # set local DDC location
    path = os.path.join(work_drive, tools.work_drive_ddc_folder)
    print(f"setting local data cache path to {path}")
    tools.env_var_set("UE-LocalDataCachePath", path)

    # additionally, the git dependencies for source engines are cached
    # in the same place
    path = os.path.join(work_drive, tools.work_drive_git_depends)
    print(f"setting git dependencies cache path to {path}")
    value = f"--cache={path}"
    tools.env_var_set("UE-UE_GITDEPS_ARGS", value)


def register_uproject_handler(
    engine_root=None, handler=None, user=True, verify=True, test=False
):
    # the engine contains a special execution to act as a shell handler
    # for files with the .uproject extension
    reg_root = winreg.HKEY_CURRENT_USER if user else winreg.HKEY_LOCAL_MACHINE

    cmdhandler = None
    if not handler:
        handler = os.path.join(
            engine_root, "Engine", "Binaries", "Win64", "UnrealVersionSelector.exe"
        )
        handler = os.path.abspath(handler)
        cmdhandler = os.path.join(
            engine_root, "Engine", "Binaries", "Win64", "UnrealVersionSelector-Cmd.exe"
        )
        handler = [handler]
        if not os.path.exists(cmdhandler):
            cmdhandler = None
        else:
            cmdhandler = [cmdhandler]
    else:
        # we have been given a custom handler,
        if not isinstance(handler, list):
            handler = [handler]
    handler[0] = os.path.abspath(handler[0])

    # we need to find an icon for the handler
    paxoicon = os.path.abspath(os.path.join(os.path.dirname(__file__), "paxo.ico"))
    iconhandler = handler[0]
    if not os.path.exists(handler[0]):
        if verify:
            raise click.ClickException(f"Could not find a handler at {handler[0]}")
        # use something else for the icon
        iconhandler = paxoicon
    # or, if it is not UnrealVersionSelector, use the icon from paxo
    if "UnrealVersion" not in os.path.basename(handler[0]):
        iconhandler = paxoicon

    cmdhandler = cmdhandler or handler

    # quoted handler for the registry
    quoted_handler = " ".join(f'"{f}"' for f in handler)
    quoted_cmdhandler = " ".join(f'"{f}"' for f in cmdhandler)

    quoted_iconhandler = f'"{iconhandler}"'
    quoted_paxoicon = f'"{paxoicon}"'

    # we must now find the appropriate place in the registry and add it.abs
    # this is a bit tricky, but we can use the python winreg module
    key = Key(reg_root, "SOFTWARE\\Classes\\.uproject")
    with key.create():
        key[""] = "Unreal.ProjectFile"

    with Key(reg_root, "SOFTWARE\\Classes\\Unreal.ProjectFile", create=True) as key:
        # we could well clear this subtree first and write it all fresh
        key[""] = "Unreal Engine Project File"
        key["VersionSelectorExecutable"] = quoted_handler
        key["VersionSelectorExecutableCmd"] = quoted_cmdhandler

        # the DefaultIcon subkey
        with key.create("DefaultIcon") as subkey:
            subkey[""] = quoted_iconhandler

        with key.create("shell") as subkey:
            with subkey.create("open") as subsubkey:
                subsubkey[""] = "Open"
                with subsubkey.create("command") as subsubsubkey:
                    subsubsubkey[""] = quoted_handler + ' /editor "%1"'

            with subkey.create("run") as subsubkey:
                subsubkey[""] = "Launch game"
                with subsubkey.create("command") as subsubsubkey:
                    subsubsubkey[""] = quoted_handler + ' /game "%1"'

            with subkey.create("rungenproj") as subsubkey:
                subsubkey[""] = "Generate Visual Studio project files"
                with subsubkey.create("command") as subsubsubkey:
                    subsubsubkey[""] = quoted_handler + ' /projectfiles "%1"'

            with subkey.create("switchversion") as subsubkey:
                subsubkey[""] = "Switch Unreal Engine version..."
                with subsubkey.create("command") as subsubsubkey:
                    subsubsubkey[""] = quoted_handler + ' /switchversion "%1"'

            with subkey.create("test") as subsubkey:
                subsubkey[""] = "run the user test thingie..."
                with subsubkey.create("command") as subsubsubkey:
                    subsubsubkey[""] = quoted_handler + ' /test "%1"'
                with subsubkey.create("icon") as subsubsubkey:
                    subsubsubkey[""] = quoted_paxoicon

            with subkey.create("hotdog") as subsubkey:
                subsubkey[""] = "Get some sausages..."
                with subsubkey.create("command") as subsubsubkey:
                    subsubsubkey[""] = "explorer.exe https://hot-dog.org"
                with subsubkey.create("Icon") as subsubsubkey:
                    subsubsubkey[""] = quoted_paxoicon

    # If the user has manually selected something other than our extension, we need to delete it. Explorer explicitly disables set access on that keys in that folder, but we can delete the whole thing.

    user_choice = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\FileExts\\.uproject\\UserChoice"
    key = Key(winreg.HKEY_CURRENT_USER, user_choice)
    if key:
        key.delete(tree=True)
    return handler


def get_uproject_handler(user=True, verify=True):
    reg_root = winreg.HKEY_CURRENT_USER if user else winreg.HKEY_LOCAL_MACHINE

    def unquote(s):
        if s[0] == '"' and s[-1] == '"':
            s = s[1:-1]
        return s

    def check_handler(s, verify=False):
        if s[0] == '"' and s[-1] == '"':
            s = s[1:-1]
        else:
            print(f"Warning: string '{s}' is not quoted")
        if not os.path.isfile(s):
            print(f"Warning: {s} does not exist")
            if verify:
                raise click.ClickException(f"Could not find {s}")

    # we must now find the appropriate place in the registry and add it.abs
    # this is a bit tricky, but we can use the python winreg module
    with Key(reg_root, "SOFTWARE\\Classes\\.uproject") as key:
        if not key:
            return None
        cls = key[""]
        if not cls:
            return None

    with Key(reg_root, f"SOFTWARE\\Classes\\{cls}") as key:
        if not key:
            return None

        exe = key.get("VersionSelectorExecutable")
        if exe:
            return unquote(exe)

        # look for the "open" subkey
        with key.subkey(r"shell\open\command") as subkey:
            if not subkey:
                return None
            cmd = subkey[""]
            # we expect something like '"C:\Program Files\Epic Games\UE_4.26\Engine\Binaries\Win64\UnrealVersionSelector.exe" /editor "%1"'
            # split it into the exe and the args, taking care of quotes, ignoring spaces inside the quotes
            parts = cmd.split('"')
            return parts[1]


def check_uproject_handler(user=True, verify=True):
    # the engine contains a special execution to act as a shell handler
    # for files with the .uproject extension
    reg_root = winreg.HKEY_CURRENT_USER if user else winreg.HKEY_LOCAL_MACHINE

    def check_handler(s, verify=False):
        if s[0] == '"' and s[-1] == '"':
            s = s[1:-1]
        else:
            print(f"Warning: string '{s}' is not quoted")
        if not os.path.isfile(s):
            print(f"Warning: {s} does not exist")
            if verify:
                raise click.ClickException(f"Could not find {s}")

    # we must now find the appropriate place in the registry and add it.abs
    # this is a bit tricky, but we can use the python winreg module
    with Key(reg_root, "SOFTWARE\\Classes\\.uproject") as key:
        if not key:
            print(".uproject handler not registered")
        else:
            value = key[""]
            if value != "Unreal.ProjectFile":
                print(".uproject handler registered to something else")
            else:
                print(".uproject handler registered")

    with Key(reg_root, "SOFTWARE\\Classes\\Unreal.ProjectFile") as key:
        if not key:
            print("Unreal.ProjectFile handler not registered")
        else:
            print(f"Unreal.ProjectFile handler registered to {key['']!r}")
            try:
                exe = key["VersionSelectorExecutable"]
                print(f"VersionSelectorExecutable registered to {exe!r}")
                check_handler(exe, verify)
            except KeyError:
                exe = None

            with key.subkey("DefaultIcon") as subkey:
                if not subkey:
                    print("DefaultIcon not registered")
                else:
                    icon = subkey[""]
                    print(f"DefaultIcon registered to {icon!r}")
                    check_handler(icon, verify)

            with key.subkey("shell") as subkey:
                if not subkey:
                    print("shell not registered")
                else:
                    for subsubkey in subkey.subkeys():
                        with subsubkey:
                            print(
                                f"shell subkey {subsubkey.name} registered to '{subsubkey['']!r}'"
                            )
                            subsubsubkey = subsubkey.subkey("command")
                            if not subsubsubkey:
                                print(f"shell subkey {subsubkey.name} has no command")
                            else:
                                with subsubsubkey:
                                    print(
                                        f"shell subkey {subsubkey.name} command is  {subsubsubkey['']!r}"
                                    )
                            subsubsubkey = subsubkey.subkey("Icon")
                            if not subsubsubkey:
                                print(f"shell subkey {subsubkey.name} has no Icon")
                            else:
                                with subsubsubkey:
                                    print(
                                        f"shell subkey {subsubkey.name} Icon set to {subsubsubkey['']!r}"
                                    )
                                    check_handler(subsubsubkey[""], verify)

    if user:
        # If the user has manually selected something other than our extension, we need to delete it. Explorer explicitly disables set access on that keys in that folder, but we can delete the whole thing.
        user_choice = "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\FileExts\\.uproject\\UserChoice"
        key = Key(winreg.HKEY_CURRENT_USER, user_choice)
        if key:
            print(
                "User has default shell action for .uproject files.  Re-register to remove it"
            )
            key.print(tree=True)


def deregister_uproject_handler(user=True):
    # the engine contains a special execution to act as a shell handler
    # for files with the .uproject extension
    reg_root = winreg.HKEY_CURRENT_USER if user else winreg.HKEY_LOCAL_MACHINE

    Key(reg_root, "SOFTWARE\\Classes\\Unreal.ProjectFile").delete(tree=True)
    Key(reg_root, "SOFTWARE\\Classes\\.uproject").delete(tree=True)
    if user:
        Key(
            reg_root,
            "SOFTWARE\\Microsoft\\Windows\\CurrentVersion\\Explorer\\FileExts\\.uproject\\UserChoice",
        ).delete(tree=True)


def get_engine_id_from_build_info(engine_path, check=True):
    # read the engine registry name from the build_info.json file
    check = True

    if check and not is_valid_root_directory(engine_path):
        raise click.ClickException(f"Engine path {engine_path} is not valid")

    try:
        with open(os.path.join(engine_path, "build_info.json")) as f:
            info = json.load(f)
        return info["engine_id"]
    except OSError:
        return None


def register_engine(engine_path, engine_id=None, check=True):
    # read the engine registry name from the build_info.json file

    if check and not is_valid_root_directory(engine_path):
        raise click.ClickException(f"Engine path {engine_path} is not valid")

    if not engine_id:
        with open(os.path.join(engine_path, "build_info.json")) as f:
            info = json.load(f)
        engine_name = info["engine_id"]
    else:
        engine_name = engine_id

    # now open the registry key for the user
    key = Key(winreg.HKEY_CURRENT_USER, "SOFTWARE\\Epic Games\\Unreal Engine\\Builds")
    with key.create():
        key[engine_name] = engine_path


def deregister_engine(engine_path=None, engine_id=None):
    if not engine_path and not engine_id:
        raise ValueError("Must specify either engine_path or engine_id")
    if not engine_id:
        assert not engine_id
        with open(os.path.join(engine_path, "build_info.json")) as f:
            info = json.load(f)
        engine_id = info["engine_id"]
    with Key(
        winreg.HKEY_CURRENT_USER,
        "SOFTWARE\\Epic Games\\Unreal Engine\\Builds",
        write=True,
    ) as key:
        if key:
            del key[engine_id]


def register_engine_old(engine_path):
    # we have a special engine registration tool in the engine folder
    tool = os.path.join(engine_path, "build-tools", "register_engine.cmd")
    subprocess.check_call([tool])


def install_prerequisites(engine_path):
    # we have a special engine registration tool in the engine folder
    tool = os.path.join(
        engine_path, "Engine", "Extras", "Redist", "en-us", "UE4PrereqSetup_x64.exe"
    )
    subprocess.check_call([tool, "/quiet"])


def start_editor(engine_path, project_path):
    # find the .uproject file in project_path
    uproject = glob.glob(os.path.join(project_path, "*.uproject"))[0]
    uproject = os.path.abspath(uproject)

    # find the ue executable
    ue = os.path.join(engine_path, "Engine", "Binaries", "Win64", "UnrealEditor.exe")

    # start the editor
    subprocess.check_call([ue, uproject])


# uproject engine registration stuff


def read_uproject(uproject):
    # read the .uproject file
    with open(uproject) as f:
        info = json.load(f)
    return info


def write_uproject(uproject, info):
    # write the .uproject file
    with open(uproject, "w") as f:
        json.dump(info, f, indent="\t")


def get_engine_association(uproject):
    # read the .uproject file
    info = read_uproject(uproject)
    engine = info.get("EngineAssociation", None)
    return engine


def set_engine_association(uproject, engine_identifier):
    # read the .uproject file
    info = read_uproject(uproject)
    if "EngineAssociation" in info:
        info["EngineAssociation"] = engine_identifier
    else:
        # recreate the dict, putting EngineAssociation at position 1 (after FileVersion)
        newinfo = {}
        for i, (key, value) in enumerate(info.items()):
            newinfo[key] = value
            if i == 0:
                newinfo["EngineAssociation"] = engine_identifier
        info = newinfo
    write_uproject(uproject, info)


def find_engine_for_project(uproject):
    """Finds the root of the engine to use for this uproject file"""
    engine_id = get_engine_association(uproject)
    if not engine_id:
        # this is a unified build, so we need to find the engine in
        # the parents
        return find_engine_in_parents(uproject)
    else:
        # guids are not engine paths, bypass the engine path test
        if not is_ue_engine_guid(engine_id):
            # first, check if the id is in fact a file path.
            # does it look like a path?  If it is not an absolute path, try to
            # use it relative to the project root
            engine_path = engine_id
            if not os.path.isabs(engine_path):
                engine_path = os.path.join(os.path.dirname(uproject), engine_id)
            if is_valid_root_directory(engine_path):
                return engine_path

        # no, not an engine path.  It must be an id, then.
        return look_up_engine(engine_id)


def is_ue_engine_guid(guid):
    """Check if this is a valid engine guid"""
    # unreal guids are 38 chars, hex digits, with dashes
    if len(guid) != 38:
        return False
    if not guid.startswith("{") and not guid.endswith("}"):
        return False
    return True


def find_engine_in_parents(uproject):
    """Given an uproject, search up the hierarhcy until an engine is found"""
    uproject = os.path.abspath(uproject)
    project_folder = os.path.dirname(uproject)
    current = os.path.dirname(project_folder)
    last = ""
    # loop until we can reach no higher
    while current != last:
        last = current
        # check if this folder contains an engine
        if is_valid_root_directory(current):
            # found the engine
            return current
        # go up a level
        current = os.path.dirname(current)


def look_up_engine(engine_id):
    """Look up the engine path from the registry"""
    engines = desktop.platform.enumerate_engine_installations()
    return engines.get(engine_id, None)


def get_editor_path(engine_path):
    """Get the path to the editor executable"""
    editor = os.path.join(
        engine_path, "Engine", "Binaries", "Win64", "UnrealEditor.exe"
    )
    if os.path.isfile(editor):
        return editor


def get_editor_from_uproject(uproject):
    """Get the path to the editor executable from the uproject file"""
    engine_path = find_engine_for_project(uproject)
    if not engine_path:
        return None
    return get_editor_path(engine_path)


"""
# project file generation
def generate_project_file(engine_dir, uproject, warn, log_file_path):
{
	arguments = " -projectfiles"

    project_directory = ProjectDirectory.get(engine_dir)

	# Build the arguments to pass to UBT. If it's a non-foreign project, just build full project files.
	if ( uproject and  project_directory.is_foreign_project(uproject) ):
		# Figure out whether it's a foreign project
		const FUProjectDictionary &ProjectDictionary = GetCachedProjectDictionary(RootDir);
		if(ProjectDictionary.IsForeignProject(ProjectFileName))
		{
			Arguments += FString::Printf(TEXT(" -project=\"%s\""), *IFileManager::Get().ConvertToAbsolutePathForExternalAppForRead(*ProjectFileName));

			// Always include game source
			Arguments += TEXT(" -game");

			// Determine whether or not to include engine source
			if(IsSourceDistribution(RootDir))
			{
				Arguments += TEXT(" -engine");
			}
			else
			{
				// If this is used within UnrealVersionSelector then we still need to pass
				// -rocket to deal with old versions that don't use Rocket.txt file
				Arguments += TEXT(" -rocket");
			}
		}
	}
	Arguments += TEXT(" -progress");

	if (!LogFilePath.IsEmpty())
	{
		Arguments += FString::Printf(TEXT(" -log=\"%s\""), *LogFilePath);
	}

	// Compile UnrealBuildTool if it doesn't exist. This can happen if we're just copying source from somewhere.
	bool bRes = true;
	Warn->BeginSlowTask(LOCTEXT("GeneratingProjectFiles", "Generating project files..."), true, true);
	if(!FPaths::FileExists(GetUnrealBuildToolExecutableFilename(RootDir)))
	{
		Warn->StatusUpdate(0, 1, LOCTEXT("BuildingUBT", "Building UnrealBuildTool..."));
		bRes = BuildUnrealBuildTool(RootDir, *Warn);
	}
	if(bRes)
	{
		Warn->StatusUpdate(0, 1, LOCTEXT("GeneratingProjectFiles", "Generating project files..."));
		bRes = RunUnrealBuildTool(LOCTEXT("GeneratingProjectFiles", "Generating project files..."), RootDir, Arguments, Warn);
	}
	Warn->EndSlowTask();
	return bRes;
}
"""



import os.path
import subprocess

import click

from . import config, p4, tools, ue
from .p4 import p4 as p4_grp
from .py import python as python_grp
from .ue import ue as ue_grp
from .uvs import uvs as uvs_grp
from .vstudio import vstudio as vstudio_grp


@click.group()
@click.option("--verbose", "-v", is_flag=True, help="Enables verbose mode.")
@click.pass_context
def cli(ctx, verbose):
    ctx.ensure_object(dict)
    ctx.obj["verbose"] = verbose


@cli.command()
def initial_setup():
    click.echo("Welcome to the initial setup of paxo.")
    click.echo(
        """
    Currently this is not automated.  Run the following commands:
    - paxo p4 install
    - paxo p4 setup

    Then, run the following to sync all depots:
    - paxo p4 sync

    Then, set up various UE things:
    - paxo ue setup

    """
    )


@cli.group()
def self():
    """work with paxo itself."""
    pass


@self.command()
def update():
    """Update paxo."""
    # rye manages our paxo installation
    # it doesn't have an "update" command, so we uninstall and install
    subprocess.run(["rye", "tools", "install", "-f", "mainframe-paxo"], check=True)
    # subprocess.run("pipx upgrade mainframe-paxo", shell=True, check=True)
    print("paxo updated, now running post_update actions...")
    subprocess.run("paxo self post-update", shell=True, check=True)


@self.command()
def post_update():
    """Run actions to refresh settings after updating paxo."""
    print("welcome to post_update_paxo")


@cli.group()
def location():
    """work with current location"""
    pass


@location.command()
@click.option("--verbose", "-v", is_flag=True, help="Enables verbose mode.")
def list(verbose):
    """list the available locations."""
    print("Available locations:")
    if not verbose:
        for location in tools.locations:
            print(f" - {location}")
    else:
        for location, info in tools.locations.items():
            print(f" - {location}")
            for key, value in info.items():
                print(f"   - {key}: {value}")


@location.command("set")
@click.option(
    "--location",
    prompt="Location",
    type=click.Choice(list(config.locations.keys())),
    default=None,
)
def location_set(location):
    """set the location."""
    p4.set_location(location)
    ue.set_location(location)
    tools.location_set(location)
    print(f"Location set to {location}")


@location.command("show")
@click.pass_context
def location_show(ctx):
    """show the current location."""
    loc = tools.location_get(empty_ok=True)
    if not loc:
        print("No location set.  Did you run initial-setup?")
        return
    if ctx.obj["verbose"]:
        print(f"Current location: {loc}")
        for key, value in tools.locations[loc].items():
            print(f" - {key}: {value}")
    else:
        print(loc)


@cli.command()
@click.option("--drive", type=click.Path(exists=False))
def subst_drive(drive):
    """Select or display the subst drive."""
    if drive:
        drive = tools.validate_drivename(drive, check_exists=False)
        tools.subst_drive_set(drive)
    else:
        drive = tools.subst_drive_get()
        print(f"Current subst drive: {drive}")
    print(f"Using drive {drive} as subst drive for {config.work_drive_dev_folder}")
    if not os.path.isfile(drive):
        print("currently not subst")


@cli.command()
@click.option("--drive", type=click.Path(exists=True, file_okay=False, dir_okay=True))
def work_drive(drive):
    """Select or display the work drive."""
    if drive:
        drive = tools.validate_drivename(drive, check_exists=True)
        tools.workdriv_set(drive)
    else:
        drive = tools.workdrive_get(empty_ok=True)
        if not drive:
            print("No work drive set.")
            return
    print(f"Using drive {drive} as work drive.")


cli.add_command(p4_grp)
cli.add_command(ue_grp)
cli.add_command(python_grp)
cli.add_command(vstudio_grp)
cli.add_command(uvs_grp)


paxo = cli
if __name__ == "__main__":
    tools.click_main(cli, obj={})



import os
import subprocess
import sys
import tkinter.messagebox

import click

from . import config
from .registry import Key

# some common environment tools for us


def validate_drivename(drivename, check_exists=True):
    drivename = drivename.upper().strip()
    if not drivename.endswith(":"):
        drivename += ":"
    if check_exists:
        if not os.path.isdir(os.path.join(drivename, "\\")):
            raise ValueError(f"Drive {drivename} does not exist")
    return drivename


# workdrive is the pyhysical drive we use for storage.
# we then use subst to map the paxdei_dev folder to a different drive letter
# typically W:  This is the subst drive.


def workdrive_set(drivename):
    drivename = validate_drivename(drivename)
    env_var_set("PD_WORKDRIVE", drivename)


def workdrive_get(empty_ok=False):
    drivename = env_var_get("PD_WORKDRIVE")
    if not drivename and not empty_ok:
        raise ValueError("PD_WORKDRIVE not set.  Did you run initial-setup?")
    return validate_drivename(drivename, check_exists=True)


# work location related stuff
def location_set(location):
    if location not in config.locations.keys():
        raise ValueError(f"Unknown location {location}")
    env_var_set("PD_LOCATION", location)


def location_get(empty_ok=False):
    location = env_var_get("PD_LOCATION")
    if not location:
        if empty_ok:
            return None
        raise ValueError("PD_LOCATION not set.  Did you run initial-setup?")
    if location not in config.locations.keys():
        raise ValueError(f"Unknown location {location}")
    return location


# setting and getting environment variables
def env_var_get(name):
    if name in os.environ:
        return os.environ[name]

    # the env var may have been set in a previous session
    # and not yet updated in _out_ environment. so we look
    # in the registry.
    with Key.current_user("Environment") as key:
        value = key.get(name, None)
        if value:
            return value
    # try the system environment
    with Key.local_machine(
        "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment"
    ) as key:
        return key.get(name, None)


def env_var_set(name, value, permanent=True, system=False):
    if permanent:
        cmd = ["setx", name, value]
        if system:
            cmd.append("/m")
        subprocess.run(cmd, check=True, capture_output=True)
    os.environ[name] = value


def env_var_del(name, permanent=True, system=False):
    if not permanent:
        try:
            del os.environ[name]
        except KeyError:
            pass
    if not permanent:
        return

    if system:
        key = Key.current_user("Environment")
    else:
        key = Key.local_machine(
            "SYSTEM\\CurrentControlSet\\Control\\Session Manager\\Environment"
        )
    with key:
        try:
            del key[name]
        except KeyError:
            pass


def addpath(path, permanent=False, system=False, infront=False):
    if infront:
        path = path + ";" + os.environ["PATH"]
    else:
        path = os.environ["PATH"] + ";" + path
    env_var_set("PATH", path, permanent, system)


# Subst!  This is a key feature of paxo.
# The user selects a a work drive, e.g. D:, with plenty of room.
# We then ensure that it has a folder called D:\paxdei_dev, and
# we subst that to P:\


def subst_drive(work_drive=None, force=False):
    """Subst the paxdei root folder to the subst-drive drive"""
    # W will have the structure
    # W:\.paxo   # this is a marker file to identify it.
    # W:\paxdei
    # W:\UE
    # W:\otherstuff

    work_drive = work_drive or workdrive_get(empty_ok=False)
    drive = subst_drive_get()

    if subst_drive == "P:":
        raise ValueError(
            "Drive P: is reserved for Pipeline.  Please select another drive."
        )
    if work_drive == drive:
        raise ValueError(
            "You must have a different drive letter for subst drive and work drive.  Call for help."
        )

    # ensure the src folder exists
    src = os.path.join(work_drive, config.work_drive_dev_folder)
    os.makedirs(src, exist_ok=True)

    # ensure that it contains the .paxo file
    with open(os.path.join(src, ".paxo"), "w") as f:
        f.write("This is the paxo work drive")

    out = subprocess.run(["subst", drive, src], capture_output=True)
    if out.returncode == 0:
        workdrive_set(work_drive)
        return True

    if "already" not in out.stdout.decode():
        out.check_returncode()

    if force:
        subprocess.run(["subst", drive, "/d"], check=True)
        subprocess.run(["subst", drive, src], check=True)
        workdrive_set(work_drive)
        return True
    return False


def subst_drive_check(drive):
    """Check if drive is a subst drive"""
    drive = drive or subst_drive_get()
    drive = validate_drivename(drive, check_exists=False)

    # if drive does not exist, we are ok
    if not os.path.isdir(drive):
        return True

    # if drive exists, we need to check if it is a subst drive
    target = is_subst(drive)
    if not target:
        raise click.ClickException(
            f"Intended subst drive {drive} exists, but is not a subst drive.  Please run 'paxo subst-drive' to select another."
        )

    # look for the .paxo file
    if not os.path.isfile(os.path.join(drive, ".paxo")):
        raise click.ClickException(
            f"Intended subst drive {drive} exists, but is not a paxo subst drive.  Please run paxo 'subst-drive' to select another."
        )

    return True


def subst_drive_get():
    drive = env_var_get("PD_SUBST_DRIVE")
    if not drive:
        drive = config.subst_drive_name
        subst_drive_set(drive)
    return validate_drivename(drive, check_exists=False)


def subst_drive_set(drive):
    drive = validate_drivename(drive, check_exists=False)
    subst_drive_check(drive)
    env_var_set("PD_SUBST_DRIVE", drive)


def is_subst(drive):
    """Check if drive is a subst drive"""
    drive = validate_drivename(drive, check_exists=False)
    out = subprocess.run(["subst"], capture_output=True)
    for line in out.stdout.decode().splitlines():
        if line.startswith(drive):
            # return the part on the right side of the subst
            return line.split("=>")[1].strip()


def click_main(command, **extra_args):
    """run a click executable with an optional exception handler for no-console"""
    # do an initial check for --gui flags here because they have not been parsed yet
    if "--gui" in sys.argv[1:]:
        gui = True
    elif "--no-gui" not in sys.argv[1:]:
        gui = False
    else:
        gui = sys.executable.endswith("pythonw.exe")
    if gui:
        try:
            return command.main(standalone_mode=False, **extra_args)
        except click.Abort:
            tkinter.messagebox.showinfo("Aborted", "Operation aborted.")
            sys.exit(1)
        except click.ClickException as e:
            tkinter.messagebox.showerror("Error", e.format_message())
            sys.exit(e.exit_code)
        except Exception as e:
            tkinter.messagebox.showerror("Error", str(e))
            raise
    else:
        return command(**extra_args)



from __future__ import annotations

import contextlib
import functools
import glob
import json
import os
import os.path
import queue
import re
import shlex
import subprocess
import threading
import time
import uuid
import weakref
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Dict, Generator, Iterable, List, Optional, Self, Tuple

import click

from . import config_cache, paths
from . import desktop_win as platform

PLATFORM_WINDOWS = True

# stuff which helps with the engine management on the desktop. Based on the
# DesktopPlatformBase class from the UnrealEngine.


def normalize_root(path: str) -> str:
    return os.path.normcase(os.path.normpath(path))


def is_guid(s: str) -> bool:
    """Check if the string is a guid"""
    try:
        uuid.UUID(hex=s)
        return True
    except ValueError:
        return False


def get_engine_descriptions() -> Iterable[Tuple[str, str, str]]:
    """Get the descriptions of the engines installed on the system."""
    for engine_id, root_dir in platform.enumerate_engine_installations().items():
        yield (engine_id, root_dir, get_engine_description(engine_id, root_dir))


def get_real_path(path):
    name = "%s[%s]" % (path[:-1], path[-1])
    found = glob.glob(name)
    if found:
        return found[0]
    return path


def get_engine_description(identifier: str, root_dir=None, format=False) -> str:
    """Get the description of the engine from the identifier."""
    if is_stock_engine_release(identifier):
        if format:
            return click.style(identifier, fg="green")
        return identifier

    # Otherwise get the path
    root_dir = root_dir or get_engine_root_dir_from_identifier(identifier)
    if not root_dir:
        return ""

    platform_root_dir = os.path.normpath(root_dir)

    # source build
    if is_source_distribution(root_dir):
        prefix = "Source"
    else:
        prefix = "Binary"
    if format:
        return f"{click.style(prefix, fg='cyan')} build at {click.style(platform_root_dir, fg='green')} "
    else:
        return f"{prefix} build at {platform_root_dir}"


def is_stock_engine_release(identifier: str) -> bool:
    """Check if the identifier is a stock engine release."""
    try:
        uuid.UUID(identifier)
        return False
    except ValueError:
        pass

    # classic UE uses only uuids for non-stock engines.  But we allow for
    # other engines, so only if this is a x.y version number do we consider it stock
    match = re.match(r"(\d+\.\d+)", identifier)
    return bool(match)


def get_engine_root_dir_from_identifier(engine_id: str) -> Optional[str]:
    """Look up the engine path from the registry"""
    engines = platform.enumerate_engine_installations()
    return engines.get(engine_id, None)


def get_engine_identifier_from_root_dir(root_dir: str) -> Optional[str]:
    root_dir = normalize_root(root_dir)
    engines = platform.enumerate_engine_installations()

    # we prefer GUID keys to custom keys.
    found = None
    for k, v in engines.items():
        if normalize_root(v) == root_dir:
            if is_guid(k):
                return k
            if not found:
                found = k
    if found:
        return found

    # otherwise, just add it
    return platform.register_engine_installation(root_dir)


def is_valid_root_directory(engine_root):
    """See if this is a proper engine root"""

    # 1 there needs to be an Engine/Binaries folder
    if not os.path.isdir(os.path.join(engine_root, "Engine", "Binaries")):
        return False
    # 2 Also check there's an Engine\Build directory.
    # This will filter out anything that has an engine-like directory structure
    # but doesn't allow building code projects - like the launcher.
    if not os.path.isdir(os.path.join(engine_root, "Engine", "Build")):
        return False

    # 3 Check for a Build.version file.  This will rule out empty directory structures
    if not os.path.isfile(
        os.path.join(engine_root, "Engine", "Build", "Build.version")
    ):
        return False

    # else, we are ok
    return True


def is_source_distribution(engine_root):
    """Check if this is a source distribution"""
    filename = os.path.join(engine_root, "Engine", "Build", "SourceDistribution.txt")
    return os.path.isfile(filename)


def is_perforce_build(engine_root):
    """Check if this is a perforce build"""
    filename = os.path.join(engine_root, "Engine", "Build", "PerforceBuild.txt")
    return os.path.isfile(filename)


def is_preferred_identifier(identifier, other_identifier):
    """Used to sort identifiers for preference"""
    m = re.match(r"(\d+\.\d+)", identifier)
    version = float(m.group(1)) if m else 0.0
    m = re.match(r"(\d+\.\d+)", other_identifier)
    other_version = float(m.group(1)) if m else 0.0
    if version != other_version:
        return version > other_version
    return identifier > other_identifier


def sort_identifiers(identifiers):
    """Sort the identifiers in a preferred order"""
    return sorted(identifiers, key=functools.cmp_to_key(is_preferred_identifier))


class ProjectDirectory:
    cache: Dict[str, ProjectDirectory] = {}

    def __init__(self, root):
        self.root = root
        # project dictionary for a given engine
        roots, projects = self.get_project_dictionary(root)
        self.project_root_dirs = roots
        self.projects = projects

    @classmethod
    def get_project_dictionary(cls, engine_root):
        roots = []
        engine_root = os.path.abspath(engine_root)
        for file in glob.glob(os.path.join(engine_root, "*.uprojectdirs")):
            with open(file) as f:
                for line in f:
                    line = line.strip()
                    if not line or line.startswith(";"):
                        continue

                    # convert to absolute path
                    if not os.path.isabs(line):
                        line = os.path.join(engine_root, line)
                    line = os.path.normcase(os.path.abspath(line))

                    # is it under the root
                    if line.startswith(os.path.normcase(engine_root)):
                        roots.append(line)
                    else:
                        print(
                            f"Warning: project search path {line} is not under engine root"
                        )

        # Search for all the projects under each root directory
        projects = []
        for root in roots:
            uprojects = glob.glob(os.path.join(root, "*", "*.uproject"))
            for up in uprojects:
                projects.append(
                    (os.path.basename(up), os.path.normcase(os.path.abspath(up)))
                )
        return roots, projects

    @classmethod
    def get(cls, engine_root):
        engine_root = os.path.abspath(engine_root)
        if engine_root in cls.cache:
            return cls.cache[engine_root]

        # create a new one
        result = cls(engine_root)
        cls.cache[engine_root] = result
        return result

    def is_foreign_project(self, project_file_name):
        # a foreign project is one which is not 'native', i.e. outside the engine root
        # check if this project is within the engine
        project_file_name = os.path.normcase(os.path.abspath(project_file_name))
        projects = [x[1] for x in self.projects]
        if project_file_name in projects:
            return False
        # could be a new project, check if its parent dir is in the project root dirs
        project_root_dir = os.path.dirname(project_file_name)
        if project_root_dir in self.project_root_dirs:
            return False

        # ok, must be foreign, then
        return True


# uproject related stuff


def get_engine_identifier_for_project(uproject):
    engine_id = get_engine_association(uproject)
    if not engine_id:
        return None

    if is_guid(engine_id):
        return engine_id

    # If identifier doesn't look like a guid, consider it a relative path to test
    # against current dir, and all parent dirs.
    # Absolute paths are also allowed and checked only once
    folder = os.path.dirname(uproject)
    lastfolder = ""
    while folder and folder != lastfolder:
        is_abs = os.path.abspath(engine_id)
        check = os.path.join(folder, engine_id) if not is_abs else engine_id
        if is_valid_root_directory(check):
            id = get_engine_identifier_from_root_dir(check)
            # remove any trailing slashes from the id, which might have been added
            # to indicate that it was a relative path
            return id.rstrip("\\/")
        if is_abs:
            break
        lastfolder = folder
        folder = os.path.dirname(folder)

    return engine_id.rstrip("\\/")


def set_engine_identifier_for_project(uproject, engine_identifier):
    # read the .uproject file
    assert engine_identifier
    info = read_uproject(uproject)
    if "EngineAssociation" in info:
        info["EngineAssociation"] = engine_identifier
    else:
        # recreate the dict, putting EngineAssociation at position 1 (after FileVersion)
        newinfo = {}
        for i, (key, value) in enumerate(info.items()):
            newinfo[key] = value
            if i == 0:
                newinfo["EngineAssociation"] = engine_identifier
        info = newinfo
    try:
        write_uproject(uproject, info)
        return True
    except OSError:
        return False


def get_engine_association(uproject):
    # read the .uproject file
    info = read_uproject(uproject)
    engine = info.get("EngineAssociation", None)
    return engine


def read_uproject(uproject):
    # read the .uproject file
    with open(uproject) as f:
        info = json.load(f)
    return info


def write_uproject(uproject, info):
    # write the .uproject file
    with open(uproject, "w") as f:
        json.dump(info, f, indent="\t")


@dataclass
class ProjectStatus:
    Name: str
    description: str
    category: str
    code_based_project: bool
    signed_sample_project: bool
    # requires_update: bool
    target_platforms: List[str]


def query_status_for_project(uproject: str) -> ProjectStatus:
    # read the .uproject file
    info = read_uproject(uproject)
    return ProjectStatus(
        Name=info.get("Name", ""),
        description=info.get("Description", ""),
        category=info.get("Category", ""),
        code_based_project=len(info.get("Modules", [])) > 0,
        signed_sample_project=info.get("IsSigned", False),
        # requires_update=info["FileVersion"] < EProjectDescriptorVersion.Latest,
        target_platforms=info.get("TargetPlatforms", []),
    )


def generate_project_files(
    root_dir: str, uproject: str, context: FeedbackContext, logfilepath: str
) -> bool:
    # Generate project files for the given project file

    project_dir = os.path.dirname(uproject)
    paths.set_engine_root(root_dir)
    paths.set_project_root(project_dir)

    args = ["-projectfiles"]

    # Build the arguments to pass to UBT. If it's a non-foreign project, just build full project files.

    # Figure out whether it's a foreign project
    pdir = ProjectDirectory.get(root_dir)
    if uproject and pdir.is_foreign_project(uproject):
        args.append(f"-project={os.path.abspath(uproject)}")

        # always include game source
        args.append("-game")

        # determine whether or not to include engine source
        if is_source_distribution(root_dir):
            args.append("-engine")
        else:
            # if this is used within UnrealVersionSelector then we still need to pass
            # -rocket to deal with old versions that don't use Rocket.txt file
            args.append("-rocket")

    args.append("-progress")

    if logfilepath:
        args.append(f"-log={logfilepath}")

    # Compile UnrealBuildTool if it doesn't exist. This can happen if we're just copying source from somewhere.

    success = True
    with context.slow("Generating project files...") as slow:
        if not os.path.exists(get_unreal_build_tool_executable_filename(root_dir)):
            slow.progress_message("Building UnrealBuildTool...")
            success = build_unreal_build_tool(root_dir, slow)
        if success:
            msg = "Generating project files..."
            slow.progress_message(msg)
            success = run_unreal_build_tool(msg, root_dir, args, slow)
        return success


def get_unreal_build_tool_executable_filename(root_dir: str) -> str:
    """Get the filename of the UnrealBuildTool executable"""
    # First, try loading it from the config file:
    configpath = os.path.join(root_dir, "Engine", "Config")
    config = config_cache.load_external_ini_file(
        ini_name="Engine",
        engine_config_dir=configpath,
        source_config_dir=configpath,
        is_base_ini_name=True,
        platform="",
        force_reload=False,
        write_dest_ini=False,
    )
    if config:
        entry = config.get("PlatformPaths", "UnrealBuildTool", fallback=None)
        if entry:
            return os.path.abspath(os.path.join(root_dir, entry))

    # use default
    return os.path.abspath(
        os.path.join(
            root_dir, "Engine/Binaries/DotNET/UnrealBuildTool/UnrealBuildTool.exe"
        )
    )


def get_unreal_build_tool_project_filename(root_dir):
    path = "Engine/Source/Programs/UnrealBuildTool/UnrealBuildTool.csproj"
    return os.path.abspath(os.path.join(root_dir, path))


def build_unreal_build_tool(root_dir, warn):
    warn.log(f"Building UnrealBuildTool in {root_dir}...")

    cs_proj_location = get_unreal_build_tool_project_filename(root_dir)
    if not os.path.isfile(cs_proj_location):
        warn.log(f"Project file not found at {cs_proj_location}")
        # return False

    if PLATFORM_WINDOWS:
        compiler_exe_filename = platform.try_read_msbuild_install_path()
        if not compiler_exe_filename:
            warn.log("Couldn't find MSBuild installation; skipping.")
            # return False
            compiler_exe_filename = "wonko.exe  "

        cmdline_params = [
            "/nologo",
            "/verbosity:quiet",
            cs_proj_location,
            "/property:Configuration=Development",
            "/property:Platform=AnyCPU",
        ]
    else:
        warn.log("Unknown platform, unable to build UnrealBuildTool.")
        return False

    warn.log(f"Running: {compiler_exe_filename} {cmdline_params}")
    success, code = warn.run_process(
        "Building UnrealBuildTool", [compiler_exe_filename] + cmdline_params
    )
    if not success or code != 0:
        return False

    # If the executable appeared where we expect it, then we were successful
    exe_path = get_unreal_build_tool_executable_filename(root_dir)
    if not os.path.isfile(exe_path):
        warn.log(f"Missing {exe_path} after build")
        return False
    return True


def run_unreal_build_tool(msg, root_dir, args, warn):
    return platform.run_unreal_build_tool(msg, root_dir, args, warn)


class FeedbackSink(ABC):
    """An object which receives feedback events"""

    @abstractmethod
    def log(self, message: str, level: str = "info") -> None:
        """Log a message to the user"""

    @abstractmethod
    def progress_message(self, message: str) -> None:
        """Report progress message to the user"""

    @abstractmethod
    def progress_value(self, value: float) -> None:
        """Report progress fraction to the user"""

    @abstractmethod
    def poll(self, delay: float = 0.0) -> str:
        """Poll the user interaction.  Can be used to perform window updates
        or to check for user interaction.  Returns a string which can be used
        to perform operations, such as "cancel" or "pause".
        """


class FeedbackContext(FeedbackSink):
    """
    A feedback context is a context which can be used to provide feedback to the user
    typically via the interactive console or with a window.
    """

    def __init__(self, parent: FeedbackContext | None = None) -> None:
        if parent is None:
            self.root = weakref.proxy(self)
        else:
            self.root = parent.root
        self.parent = parent
        self.log_lines: list[tuple[str, str]] = []
        self.current_slow_task: str | None = None

    def log(self, message, level: str = "INFO") -> None:
        """Log a message to the user"""
        if self.parent:
            self.root.log(message, level)
            return
        self.log_lines.append((level, message))

    def get_logs(self, show_level: bool = False) -> str:
        if show_level:
            return "".join(f"{level}: {message}\n" for level, message in self.log_lines)
        else:
            return "".join(f"{message}\n" for level, message in self.log_lines)

    def begin_slow_task(self, message: str, **kwargs) -> None:
        assert self.current_slow_task is None
        self.current_slow_task = message
        self.progress_message(message)

    def end_slow_task(self) -> None:
        assert self.current_slow_task is not None
        self.current_slow_task = None

    @contextlib.contextmanager
    def slow(self, message: str, nested=True, **kwargs) -> Generator[Self, None, None]:
        """A context manager for slow tasks"""
        if nested:
            ctx = self.__class__(parent=self)
        else:
            ctx = self
        ctx.begin_slow_task(message, **kwargs)
        try:
            yield ctx
        finally:
            ctx.end_slow_task()

    def poll(self, delay: float = 0.0) -> str:
        """Poll the user interaction.  Can be used to perform window updates
        or to check for user interaction.  Returns a string which can be used
        to perform operations, such as "cancel" or "pause".
        """
        if delay:
            time.sleep(delay)
        return ""

    def run_process(self, message, args, **kwargs) -> Tuple[bool, int]:
        with self.slow(message) as slow:
            try:
                p = subprocess.Popen(
                    args,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                    text=True,
                    creationflags=subprocess.CREATE_NO_WINDOW,
                    **kwargs,
                )
            except OSError as e:
                slow.log(f"Could not create process {args}: {e}", level="error")
                return (False, -1)

            # now read the pipe and output each line to the slow context
            # until we don't get any more lines
            assert p.stdout is not None
            reader = NBReader(p.stdout)
            while True:
                line = reader.try_readline()
                if line == "":
                    break  # end of file
                if line is None:
                    # no data, work the io context for a bit
                    poll = self.poll(0.1)
                    if poll == "cancel":
                        p.terminate()
                        break
                    continue
                line = line.rstrip()
                if self.parse_command(line):
                    continue
                slow.log(line)

            p.wait()
            if p.returncode:
                slow.log(f"Process {args} exited with code {p.returncode}")
            return (True, p.returncode)

    def parse_command(self, line: str) -> bool:
        # search for a message of the form
        # "@progress ["msg"|'msg'] nominator denominator"
        if not line.startswith("@progress"):
            return False
        try:
            parts = shlex.split(line)
        except ValueError:
            return False
        if parts and parts[0] == "@progress":
            del parts[0]
            msg = None
            while len(parts):
                try:
                    if parts[0].endswith("%"):
                        fraction = float(parts[0][:-1]) / 100
                    elif len(parts) >= 2:
                        nominator = int(parts[0])
                        denominator = int(parts[1])
                        fraction = nominator / denominator if denominator else 0
                    self.progress_value(fraction)
                    break
                except ValueError:
                    msg = parts[0]
                    del parts[0]
            if msg:
                self.progress_message(msg)
                return True
        return False


class NBReader:
    def __init__(self, stream):
        self.stream = stream
        self.queue = queue.SimpleQueue()
        self.thread = threading.Thread(target=self.enqueue)
        self.stop = False
        self.thread.start()

    def enqueue(self):
        # read the pipe and output each line to the slow context until an
        # empty line is received, signalling the end of the process
        try:
            while not self.stop:
                line = self.stream.readline()
                if not line:
                    break
                self.queue.put(line)
        finally:
            self.queue.put("")

    def try_readline(self, timeout: float = 0.0) -> str | None:
        try:
            if timeout:
                return self.queue.get(timeout=timeout)
            return self.queue.get(block=False)
        except queue.Empty:
            return None


class ClickFeedbackContext(FeedbackContext):
    """A progress context which uses click to display progress"""

    def __init__(self, parent=None, click_ctx=None):
        super().__init__(parent=parent)
        self.click_ctx = click_ctx
        self.click_ctx = self.root.click_ctx
        self.last_msg = None

    @contextlib.contextmanager
    def slow(self, msg, *args, **kwargs):
        with click.progressbar(
            length=100, label=msg, show_eta=False, show_percent=True
        ) as bar:
            yield self.__class__(parent=self, click_ctx=bar)

    def progress_message(self, message: str) -> None:
        assert self.click_ctx is not None
        if message != self.last_msg:
            click.echo(message)
            self.last_msg = message

    def progress_value(self, fraction: float) -> None:
        self.click_ctx.update(int(fraction * 100))

    def log(self, message, level="info"):
        super().log(message, level)
        level = level.lower()
        if "error" in level:
            click.secho(message, fg="red")
        elif "warn" in level:
            click.secho(message, fg="yellow")
        else:
            click.echo(message)


class SinkFeedbackContext(FeedbackContext):
    def __init__(
        self,
        parent: SinkFeedbackContext | None = None,
        sink: FeedbackSink | None = None,
    ):
        super().__init__(parent=parent)
        self.sink: FeedbackSink
        if parent is None:
            assert sink is not None
            self.sink = sink
        else:
            self.sink = parent.sink

    def progress_message(self, message: str) -> None:
        self.sink.progress_message(message)

    def progress_value(self, value: float) -> None:
        return self.sink.progress_value(value)

    def log(self, message, level: str = "INFO") -> None:
        super().log(message, level)
        self.sink.log(message, level)

    def poll(self, delay: float = 0.0) -> str:
        return self.sink.poll(delay)
